##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Union, List, Tuple

import networkx as nx
import numpy as np

from pyquil.noise import NoiseModel
from pyquil.parameters import Parameter
from pyquil.quilatom import unpack_qubit
from pyquil.quilbase import Gate

THETA = Parameter("theta")
"Used as the symbolic parameter in RZ, CPHASE gates."

DEFAULT_QUBIT_TYPE = "Xhalves"
DEFAULT_EDGE_TYPE = "CZ"

Qubit = namedtuple("Qubit", ["id", "type", "dead"])
Edge = namedtuple("Edge", ["targets", "type", "dead"])
_ISA = namedtuple("_ISA", ["qubits", "edges"])
QubitSpecs = namedtuple("_QubitSpecs", ["id", "fRO", "f1QRB", "T1", "T2", "fActiveReset"])
EdgeSpecs = namedtuple("_QubitQubitSpecs", ["targets", "fBellState", "fCZ", "fCPHASE"])
_Specs = namedtuple("_Specs", ["qubits_specs", "edges_specs"])


class ISA(_ISA):
    """
    Basic Instruction Set Architecture specification.

    :ivar Sequence[Qubit] qubits: The qubits associated with the ISA.
    :ivar Sequence[Edge] edges: The multi-qubit gates.
    """

    def to_dict(self):
        """
        Create a JSON-serializable representation of the ISA.

        The dictionary representation is of the form::

            {
                "1Q": {
                    "0": {
                        "type": "Xhalves"
                    },
                    "1": {
                        "type": "Xhalves",
                        "dead": True
                    },
                    ...
                },
                "2Q": {
                    "1-4": {
                        "type": "CZ"
                    },
                    "1-5": {
                        "type": "CZ"
                    },
                    ...
                },
                ...
            }

        :return: A dictionary representation of self.
        :rtype: Dict[str, Any]
        """

        def _maybe_configure(o, t):
            # type: (Union[Qubit,Edge], str) -> dict
            """
            Exclude default values from generated dictionary.

            :param Union[Qubit,Edge] o: The object to serialize
            :param str t: The default value for ``o.type``.
            :return: d
            """
            d = {}
            if o.type != t:
                d["type"] = o.type
            if o.dead:
                d["dead"] = o.dead
            return d

        return {
            "1Q": {"{}".format(q.id): _maybe_configure(q, DEFAULT_QUBIT_TYPE) for q in self.qubits},
            "2Q": {"{}-{}".format(*edge.targets): _maybe_configure(edge, DEFAULT_EDGE_TYPE)
                   for edge in self.edges}
        }

    @staticmethod
    def from_dict(d):
        """
        Re-create the ISA from a dictionary representation.

        :param Dict[str,Any] d: The dictionary representation.
        :return: The restored ISA.
        :rtype: ISA
        """
        return ISA(
            qubits=sorted([Qubit(id=int(qid),
                                 type=q.get("type", DEFAULT_QUBIT_TYPE),
                                 dead=q.get("dead", False))
                           for qid, q in d["1Q"].items()],
                          key=lambda qubit: qubit.id),
            edges=sorted([Edge(targets=[int(q) for q in eid.split('-')],
                               type=e.get("type", DEFAULT_EDGE_TYPE),
                               dead=e.get("dead", False))
                          for eid, e in d["2Q"].items()],
                         key=lambda edge: edge.targets),
        )


def gates_in_isa(isa):
    """
    Generate the full gateset associated with an ISA.

    :param ISA isa: The instruction set architecture for a QPU.
    :return: A sequence of Gate objects encapsulating all gates compatible with the ISA.
    :rtype: Sequence[Gate]
    """
    gates = []
    for q in isa.qubits:
        if q.dead:
            # TODO: dead qubits may in the future lead to some implicit re-indexing
            continue
        if q.type in ["Xhalves"]:
            gates.extend([
                Gate("I", [], [unpack_qubit(q.id)]),
                Gate("RX", [np.pi / 2], [unpack_qubit(q.id)]),
                Gate("RX", [-np.pi / 2], [unpack_qubit(q.id)]),
                Gate("RX", [np.pi], [unpack_qubit(q.id)]),
                Gate("RX", [-np.pi], [unpack_qubit(q.id)]),
                Gate("RZ", [THETA], [unpack_qubit(q.id)]),
            ])
        else:  # pragma no coverage
            raise ValueError("Unknown qubit type: {}".format(q.type))

    for e in isa.edges:
        if e.dead:
            continue
        targets = [unpack_qubit(t) for t in e.targets]
        if e.type in ["CZ", "ISWAP"]:
            gates.append(Gate(e.type, [], targets))
            gates.append(Gate(e.type, [], targets[::-1]))
        elif e.type in ["CPHASE"]:
            gates.append(Gate(e.type, [THETA], targets))
            gates.append(Gate(e.type, [THETA], targets[::-1]))
        else:  # pragma no coverage
            raise ValueError("Unknown edge type: {}".format(e.type))
    return gates


class Specs(_Specs):
    """
    Basic specifications for the device, such as gate fidelities and coherence times.

    :ivar List[QubitSpecs] qubits_specs: The specs associated with individual qubits.
    :ivar List[EdgesSpecs] edges_specs: The specs associated with edges, or qubit-qubit pairs.
    """

    def f1QRBs(self):
        """
        Get a dictionary of single-qubit randomized benchmarking fidelities (normalized to unity)
        from the specs, keyed by qubit index.

        :return: A dictionary of 1QRBs, normalized to unity.
        :rtype: Dict[int, float]
        """
        return {qs.id: qs.f1QRB for qs in self.qubits_specs}

    def fROs(self):
        """
        Get a dictionary of single-qubit readout fidelities (normalized to unity)
        from the specs, keyed by qubit index.

        :return: A dictionary of RO fidelities, normalized to unity.
        :rtype: Dict[int, float]
        """
        return {qs.id: qs.fRO for qs in self.qubits_specs}

    def fActiveResets(self):
        """
        Get a dictionary of single-qubit active reset fidelities (normalized to unity) from the
        specs, keyed by qubit index.

        :return: A dictionary of reset fidelities, normalized to unity.
        """
        return {qs.id: qs.fActiveReset for qs in self.qubits_specs}

    def T1s(self):
        """
        Get a dictionary of T1s (in seconds) from the specs, keyed by qubit index.

        :return: A dictionary of T1s, in seconds.
        :rtype: Dict[int, float]
        """
        return {qs.id: qs.T1 for qs in self.qubits_specs}

    def T2s(self):
        """
        Get a dictionary of T2s (in seconds) from the specs, keyed by qubit index.

        :return: A dictionary of T2s, in seconds.
        :rtype: Dict[int, float]
        """
        return {qs.id: qs.T2 for qs in self.qubits_specs}

    def fBellStates(self):
        """
        Get a dictionary of two-qubit Bell state fidelities (normalized to unity)
        from the specs, keyed by targets (qubit-qubit pairs).

        :return: A dictionary of Bell state fidelities, normalized to unity.
        :rtype: Dict[tuple(int, int), float]
        """
        return {tuple(es.targets): es.fBellState for es in self.edges_specs}

    def fCZs(self):
        """
        Get a dictionary of CZ fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of CZ fidelities, normalized to unity.
        :rtype: Dict[tuple(int, int), float]
        """
        return {tuple(es.targets): es.fCZ for es in self.edges_specs}

    def fCPHASEs(self):
        """
        Get a dictionary of CPHASE fidelities (normalized to unity) from the specs,
        keyed by targets (qubit-qubit pairs).

        :return: A dictionary of CPHASE fidelities, normalized to unity.
        :rtype: Dict[tuple(int, int), float]
        """
        return {tuple(es.targets): es.fCPHASE for es in self.edges_specs}

    def to_dict(self):
        """
        Create a JSON-serializable representation of the device Specs.

        The dictionary representation is of the form::

            {
                '1Q': {
                    "0": {
                        "f1QRB": 0.99,
                        "T1": 20e-6,
                        ...
                    },
                    "1": {
                        "f1QRB": 0.989,
                        "T1": 19e-6,
                        ...
                    },
                    ...
                },
                '2Q': {
                    "1-4": {
                        "fBellState": 0.93,
                        "fCZ": 0.92,
                        "fCPHASE": 0.91
                    },
                    "1-5": {
                        "fBellState": 0.9,
                        "fCZ": 0.89,
                        "fCPHASE": 0.88
                    },
                    ...
                },
                ...
            }

        :return: A dctionary representation of self.
        :rtype: Dict[str, Any]
        """
        return {
            '1Q': {
                "{}".format(qs.id): {
                    'f1QRB': qs.f1QRB,
                    'fRO': qs.fRO,
                    'T1': qs.T1,
                    'T2': qs.T2,
                    'fActiveReset': qs.fActiveReset
                } for qs in self.qubits_specs
            },
            '2Q': {
                "{}-{}".format(*es.targets): {
                    'fBellState': es.fBellState,
                    'fCZ': es.fCZ,
                    'fCPHASE': es.fCPHASE
                } for es in self.edges_specs
            }
        }

    @staticmethod
    def from_dict(d):
        """
        Re-create the Specs from a dictionary representation.

        :param Dict[str, Any] d: The dictionary representation.
        :return: The restored Specs.
        :rtype: Specs
        """
        return Specs(
            qubits_specs=sorted([QubitSpecs(id=int(q),
                                            fRO=qspecs.get('fRO'),
                                            f1QRB=qspecs.get('f1QRB'),
                                            T1=qspecs.get('T1'),
                                            T2=qspecs.get('T2'),
                                            fActiveReset=qspecs.get('fActiveReset'))
                                 for q, qspecs in d["1Q"].items()],
                                key=lambda qubit_specs: qubit_specs.id),
            edges_specs=sorted([EdgeSpecs(targets=[int(q) for q in e.split('-')],
                                          fBellState=especs.get('fBellState'),
                                          fCZ=especs.get('fCZ'),
                                          fCPHASE=especs.get('fCPHASE'))
                                for e, especs in d["2Q"].items()],
                               key=lambda edge_specs: edge_specs.targets)
        )


def isa_from_graph(graph: nx.Graph, oneq_type='Xhalves', twoq_type='CZ') -> ISA:
    """
    Generate an ISA object from a NetworkX graph.

    :param graph: The graph
    :param oneq_type: The type of 1-qubit gate. Currently 'Xhalves'
    :param twoq_type: The type of 2-qubit gate. One of 'CZ' or 'CPHASE'.
    """
    all_qubits = list(range(max(graph.nodes) + 1))
    qubits = [Qubit(i, type=oneq_type, dead=i not in graph.nodes) for i in all_qubits]
    edges = [Edge(sorted((a, b)), type=twoq_type, dead=False) for a, b in graph.edges]
    return ISA(qubits, edges)


def specs_from_graph(graph: nx.Graph):
    """
    Generate a Specs object from a NetworkX graph with placeholder values for the actual specs.

    :param graph: The graph
    """
    qspecs = [QubitSpecs(id=q, fRO=0.90, f1QRB=0.99, T1=30e-6, T2=30e-6, fActiveReset=0.99)
              for q in graph.nodes]
    especs = [EdgeSpecs(targets=(q1, q2), fBellState=0.90, fCZ=0.90, fCPHASE=0.80)
              for q1, q2 in graph.edges]
    return Specs(qspecs, especs)


def isa_to_graph(isa: ISA) -> nx.Graph:
    """
    Construct a NetworkX qubit topology from an ISA object.

    This discards information about supported gates.

    :param isa: The ISA.
    """
    return nx.from_edgelist(e.targets for e in isa.edges if not e.dead)


class AbstractDevice(ABC):

    @abstractmethod
    def qubits(self):
        """
        A sorted list of qubits in the device topology.
        """

    @abstractmethod
    def qubit_topology(self) -> nx.Graph:
        """
        The connectivity of qubits in this device given as a NetworkX graph.
        """

    @abstractmethod
    def get_isa(self, oneq_type='Xhalves', twoq_type='CZ') -> ISA:
        """
        Construct an ISA suitable for targeting by compilation.

        This will raise an exception if the requested ISA is not supported by the device.

        :param oneq_type: The family of one-qubit gates to target
        :param twoq_type: The family of two-qubit gates to target
        """

    @abstractmethod
    def get_specs(self) -> Specs:
        """
        Construct a Specs object required by compilation
        """


class Device(AbstractDevice):
    """
    A device (quantum chip) that can accept programs.

    Only devices that are online will actively be
    accepting new programs. In addition to the ``self._raw`` attribute, two other attributes are
    optionally constructed from the entries in ``self._raw`` -- ``isa`` and ``noise_model`` -- which
    should conform to the dictionary format required by the ``.from_dict()`` methods for ``ISA``
    and ``NoiseModel``, respectively.

    :ivar dict _raw: Raw JSON response from the server with additional information about the device.
    :ivar ISA isa: The instruction set architecture (ISA) for the device.
    :ivar NoiseModel noise_model: The noise model for the device.
    """

    def __init__(self, name, raw):
        """
        :param name: name of the device
        :param raw: raw JSON response from the server with additional information about this device.
        """
        self.name = name
        self._raw = raw

        # TODO: Introduce distinction between supported ISAs and target ISA
        self._isa = ISA.from_dict(raw['isa']) if 'isa' in raw and raw['isa'] != {} else None
        self.specs = Specs.from_dict(raw['specs']) if raw.get('specs') else None
        self.noise_model = NoiseModel.from_dict(raw['noise_model']) \
            if raw.get('noise_model') else None

    @property
    def isa(self):
        warnings.warn("Accessing the static ISA is deprecated. Use `get_isa`", DeprecationWarning)
        return self._isa

    def qubits(self):
        return sorted(q.id for q in self._isa.qubits if not q.dead)

    def qubit_topology(self) -> nx.Graph:
        """
        The connectivity of qubits in this device given as a NetworkX graph.
        """
        return isa_to_graph(self._isa)

    def get_isa(self, oneq_type='Xhalves', twoq_type='CZ') -> ISA:
        """
        Construct an ISA suitable for targeting by compilation.

        This will raise an exception if the requested ISA is not supported by the device.

        :param oneq_type: The family of one-qubit gates to target
        :param twoq_type: The family of two-qubit gates to target
        """
        qubits = [Qubit(id=q.id, type=oneq_type, dead=q.dead) for q in self._isa.qubits]
        edges = [Edge(targets=e.targets, type=twoq_type, dead=e.dead) for e in self._isa.edges]
        return ISA(qubits, edges)

    def get_specs(self):
        return self.specs

    def __str__(self):
        return '<Device {}>'.format(self.name)

    def __repr__(self):
        return str(self)


class NxDevice(AbstractDevice):
    """A shim over the AbstractDevice API backed by a NetworkX graph.

    A ``Device`` holds information about the physical device.
    Specifically, you might want to know about connectivity, available gates, performance specs,
    and more. This class implements the AbstractDevice API for devices not available via
    ``get_devices()``. Instead, the user is responsible for constructing a NetworkX
    graph which represents a chip topology.
    """

    def __init__(self, topology: nx.Graph) -> None:
        self.topology = topology

    def qubit_topology(self):
        return self.topology

    def get_isa(self, oneq_type='Xhalves', twoq_type='CZ'):
        return isa_from_graph(self.topology, oneq_type=oneq_type, twoq_type=twoq_type)

    def get_specs(self):
        return specs_from_graph(self.topology)

    def qubits(self) -> List[int]:
        return sorted(self.topology.nodes)

    def edges(self) -> List[Tuple[int, int]]:
        return sorted(tuple(sorted(pair)) for pair in self.topology.edges)  # type: ignore



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for working with Pauli algebras.
"""

from __future__ import division
from itertools import product
import numpy as np
import copy

from pyquil.quilatom import QubitPlaceholder

from .quil import Program
from .gates import H, RZ, RX, CNOT, X, PHASE, QUANTUM_GATES
from numbers import Number
from collections import Sequence, OrderedDict
import warnings
from six import integer_types as six_integer_types
from six.moves import range

PAULI_OPS = ["X", "Y", "Z", "I"]
PAULI_PROD = {'ZZ': 'I', 'YY': 'I', 'XX': 'I', 'II': 'I',
              'XY': 'Z', 'XZ': 'Y', 'YX': 'Z', 'YZ': 'X', 'ZX': 'Y',
              'ZY': 'X', 'IX': 'X', 'IY': 'Y', 'IZ': 'Z',
              'ZI': 'Z', 'YI': 'Y', 'XI': 'X',
              'X': 'X', 'Y': 'Y', 'Z': 'Z', 'I': 'I'}
PAULI_COEFF = {'ZZ': 1.0, 'YY': 1.0, 'XX': 1.0, 'II': 1.0,
               'XY': 1.0j, 'XZ': -1.0j, 'YX': -1.0j, 'YZ': 1.0j, 'ZX': 1.0j,
               'ZY': -1.0j, 'IX': 1.0, 'IY': 1.0, 'IZ': 1.0, 'ZI': 1.0,
               'YI': 1.0, 'XI': 1.0,
               'X': 1.0, 'Y': 1.0, 'Z': 1.0, 'I': 1.0}


class UnequalLengthWarning(Warning):
    def __init__(self, *args, **kwargs):
        super(UnequalLengthWarning, self).__init__(*args, **kwargs)


integer_types = six_integer_types + (np.int64, np.int32, np.int16, np.int8)
"""Explicitly include numpy integer dtypes (for python 3)."""

HASH_PRECISION = 1e6
"""The precision used when hashing terms to check equality. The simplify() method
uses np.isclose() for coefficient comparisons to 0 which has its own default precision. We
can't use np.isclose() for hashing terms though.
"""


def _valid_qubit(index):
    return ((isinstance(index, integer_types) and index >= 0)
            or isinstance(index, QubitPlaceholder))


class PauliTerm(object):
    """A term is a product of Pauli operators operating on different qubits.
    """

    def __init__(self, op, index, coefficient=1.0):
        """ Create a new Pauli Term with a Pauli operator at a particular index and a leading
        coefficient.

        :param string op: The Pauli operator as a string "X", "Y", "Z", or "I"
        :param int index: The qubit index that that operator is applied to.
        :param float coefficient: The coefficient multiplying the operator, e.g. 1.5 * Z_1
        """
        assert op in PAULI_OPS
        assert _valid_qubit(index)

        self._ops = OrderedDict()
        if op != "I":
            self._ops[index] = op
        if not isinstance(coefficient, Number):
            raise ValueError("coefficient of PauliTerm must be a Number.")
        self.coefficient = complex(coefficient)

    def id(self, sort_ops=True):
        """
        Returns an identifier string for the PauliTerm (ignoring the coefficient).

        Don't use this to compare terms. This function will not work with qubits that
        aren't sortable.

        :param sort_ops: Whether to sort operations by qubit. This is True by default for
            backwards compatibility but will change in pyQuil 2.0. Callers should never rely
            on comparing id's for testing equality. See ``operations_as_set`` instead.
        :return: A string representation of this term's operations.
        :rtype: string
        """
        if sort_ops and len(self._ops) > 1:
            warnings.warn("`PauliTerm.id()` will not work on PauliTerms where the qubits are not "
                          "sortable and should be avoided in favor of `operations_as_set`.",
                          FutureWarning)
            return ''.join("{}{}".format(self._ops[q], q) for q in sorted(self._ops.keys()))
        else:
            return ''.join("{}{}".format(p, q) for q, p in self._ops.items())

    def operations_as_set(self):
        """
        Return a frozenset of operations in this term.

        Use this in place of :py:func:`id` if the order of operations in the term does not
        matter.

        :return: frozenset of strings representing Pauli operations
        """
        return frozenset(self._ops.items())

    def __eq__(self, other):
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError("Can't compare PauliTerm with object of type {}.".format(type(other)))
        elif isinstance(other, PauliSum):
            return other == self
        else:
            return (self.operations_as_set() == other.operations_as_set()
                    and np.isclose(self.coefficient, other.coefficient))

    def __hash__(self):
        return hash((
            round(self.coefficient.real * HASH_PRECISION),
            round(self.coefficient.imag * HASH_PRECISION),
            self.operations_as_set()
        ))

    def __ne__(self, other):
        # x!=y and x<>y call __ne__() instead of negating __eq__
        # This is only a weirdness in python2 as in python 3 __ne__ defaults to the inversion of
        # __eq__
        return not self.__eq__(other)

    def __len__(self):
        """
        The length of the PauliTerm is the number of Pauli operators in the term. A term that
        consists of only a scalar has a length of zero.
        """
        return len(self._ops)

    def copy(self):
        """
        Properly creates a new PauliTerm, with a completely new dictionary
        of operators
        """
        new_term = PauliTerm("I", 0, 1.0)  # create new object
        # manually copy all attributes over
        for key in self.__dict__.keys():
            val = self.__dict__[key]
            if isinstance(val, (dict, list, set)):  # mutable types
                new_term.__dict__[key] = copy.copy(val)
            else:  # immutable types
                new_term.__dict__[key] = val

        return new_term

    @property
    def program(self):
        return Program([QUANTUM_GATES[gate](q) for q, gate in self])

    def get_qubits(self):
        """Gets all the qubits that this PauliTerm operates on.
        """
        return list(self._ops.keys())

    def __getitem__(self, i):
        return self._ops.get(i, "I")

    def __iter__(self):
        for i in self.get_qubits():
            yield i, self[i]

    def _multiply_factor(self, factor, index):
        new_term = PauliTerm("I", 0)
        new_coeff = self.coefficient
        new_ops = self._ops.copy()

        ops = self[index] + factor
        new_op = PAULI_PROD[ops]
        if new_op != "I":
            new_ops[index] = new_op
        else:
            del new_ops[index]
        new_coeff *= PAULI_COEFF[ops]

        new_term._ops = new_ops
        new_term.coefficient = new_coeff

        return new_term

    def __mul__(self, term):
        """Multiplies this Pauli Term with another PauliTerm, PauliSum, or number according to the
        Pauli algebra rules.

        :param term: (PauliTerm or PauliSum or Number) A term to multiply by.
        :returns: The product of this PauliTerm and term.
        :rtype: PauliTerm
        """
        if isinstance(term, Number):
            return term_with_coeff(self, self.coefficient * term)
        elif isinstance(term, PauliSum):
            return (PauliSum([self]) * term).simplify()
        else:
            new_term = PauliTerm("I", 0, 1.0)
            new_term._ops = self._ops.copy()
            new_coeff = self.coefficient * term.coefficient
            for index, op in term:
                new_term = new_term._multiply_factor(op, index)

            return term_with_coeff(new_term, new_term.coefficient * new_coeff)

    def __rmul__(self, other):
        """Multiplies this PauliTerm with another object, probably a number.

        :param other: A number or PauliTerm to multiply by
        :returns: A new PauliTerm
        :rtype: PauliTerm
        """
        assert isinstance(other, Number)
        return self * other

    def __pow__(self, power):
        """Raises this PauliTerm to power.

        :param int power: The power to raise this PauliTerm to.
        :return: The power-fold product of power.
        :rtype: PauliTerm
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")
        result = ID()

        identities = [PauliTerm('I', qubit) for qubit in self.get_qubits()]
        if not identities:
            # There weren't any nontrivial operators
            return term_with_coeff(self, 1)
        for identity in identities:
            result *= identity
        for _ in range(power):
            result *= self
        return result

    def __add__(self, other):
        """Adds this PauliTerm with another one.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the sum of this PauliTerm and other
        :rtype: PauliSum
        """
        if isinstance(other, Number):
            return self + PauliTerm("I", 0, other)
        elif isinstance(other, PauliSum):
            return other + self
        else:
            new_sum = PauliSum([self, other])
            return new_sum.simplify()

    def __radd__(self, other):
        """Adds this PauliTerm with a Number.

        :param other: A PauliTerm object or a Number
        :returns: A new PauliTerm
        :rtype: PauliTerm
        """
        assert isinstance(other, Number)
        return PauliTerm("I", 0, other) + self

    def __sub__(self, other):
        """Subtracts a PauliTerm from this one.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        :rtype: PauliSum
        """
        return self + -1. * other

    def __rsub__(self, other):
        """Subtracts this PauliTerm from a Number or PauliTerm.

        :param other: A PauliTerm object or a Number
        :returns: A PauliSum object representing the difference of this PauliTerm and term
        :rtype: PauliSum
        """
        return other + -1. * self

    def __str__(self):
        term_strs = []
        for index in self._ops.keys():
            term_strs.append("%s%s" % (self[index], index))

        if len(term_strs) == 0:
            term_strs.append("I")
        out = "%s*%s" % (self.coefficient, '*'.join(term_strs))
        return out

    @classmethod
    def from_list(cls, terms_list, coefficient=1.0):
        """
        Allocates a Pauli Term from a list of operators and indices. This is more efficient than
        multiplying together individual terms.

        :param list terms_list: A list of tuples, e.g. [("X", 0), ("Y", 1)]
        :return: PauliTerm
        """
        pterm = PauliTerm("I", 0)
        assert all([op[0] in PAULI_OPS for op in terms_list])

        indices = [op[1] for op in terms_list]
        assert all(_valid_qubit(index) for index in indices)

        # this is because from_list doesn't call simplify in order to be more efficient.
        if len(set(indices)) != len(indices):
            raise ValueError("Elements of PauliTerm that are allocated using from_list must "
                             "be on disjoint qubits. Use PauliTerm multiplication to simplify "
                             "terms instead.")

        for op, index in terms_list:
            if op != "I":
                pterm._ops[index] = op
        if not isinstance(coefficient, Number):
            raise ValueError("coefficient of PauliTerm must be a Number.")
        pterm.coefficient = complex(coefficient)
        return pterm

    def pauli_string(self, qubits=None):
        """
        Return a string representation of this PauliTerm mod its phase, as a concatenation of the string representation
        of the
        >>> p = PauliTerm("X", 0) * PauliTerm("Y", 1, 1.j)
        >>> p.pauli_string()
        "XY"
        >>> p.pauli_string([0])
        "X"
        >>> p.pauli_string([0, 2])
        "XI"

        :param list qubits: The list of qubits to represent, given as ints. If None, defaults to all qubits in this
         PauliTerm.
        :return: The string representation of this PauliTerm, modulo its phase.
        :rtype: String
        """
        qubit_term_mapping = dict(self.operations_as_set())
        if qubits is None:
            qubits = [qubit for qubit, _ in qubit_term_mapping.items()]
        ps = ""
        for qubit in qubits:
            try:
                ps += qubit_term_mapping[qubit]
            except KeyError:
                ps += "I"
        return ps


# For convenience, a shorthand for several operators.
def ID():
    """
    The identity Pauli Term.
    """
    return PauliTerm("I", 0, 1)


def ZERO():
    """
    The zero Pauli Term.
    """
    return PauliTerm("I", 0, 0)


def sI(q):
    """
    A function that returns the identity operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("I", q)


def sX(q):
    """
    A function that returns the sigma_X operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("X", q)


def sY(q):
    """
    A function that returns the sigma_Y operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("Y", q)


def sZ(q):
    """
    A function that returns the sigma_Z operator on a particular qubit.

    :param int qubit_index: The index of the qubit
    :returns: A PauliTerm object
    :rtype: PauliTerm
    """
    return PauliTerm("Z", q)


def term_with_coeff(term, coeff):
    """
    Change the coefficient of a PauliTerm.

    :param PauliTerm term: A PauliTerm object
    :param Number coeff: The coefficient to set on the PauliTerm
    :returns: A new PauliTerm that duplicates term but sets coeff
    :rtype: PauliTerm
    """
    if not isinstance(coeff, Number):
        raise ValueError("coeff must be a Number")
    new_pauli = term.copy()
    # We cast to a complex number to ensure that internally the coefficients remain compatible.
    new_pauli.coefficient = complex(coeff)
    return new_pauli


class PauliSum(object):
    """A sum of one or more PauliTerms.
    """

    def __init__(self, terms):
        """
        :param Sequence terms: A Sequence of PauliTerms.
        """
        if not (isinstance(terms, Sequence)
                and all([isinstance(term, PauliTerm) for term in terms])):
            raise ValueError("PauliSum's are currently constructed from Sequences of PauliTerms.")
        if len(terms) == 0:
            self.terms = [0.0 * ID()]
        else:
            self.terms = terms

    def __eq__(self, other):
        """Equality testing to see if two PauliSum's are equivalent.

        :param PauliSum other: The PauliSum to compare this PauliSum with.
        :return: True if other is equivalent to this PauliSum, False otherwise.
        :rtype: bool
        """
        if not isinstance(other, (PauliTerm, PauliSum)):
            raise TypeError("Can't compare PauliSum with object of type {}.".format(type(other)))
        elif isinstance(other, PauliTerm):
            return self == PauliSum([other])
        elif len(self.terms) != len(other.terms):
            warnings.warn(UnequalLengthWarning("These PauliSums have a different number of terms."))
            return False

        return set(self.terms) == set(other.terms)

    def __ne__(self, other):
        """Inequality testing to see if two PauliSum's are not equivalent.

        :param PauliSum other: The PauliSum to compare this PauliSum with.
        :return: False if other is equivalent to this PauliSum, True otherwise.
        :rtype: bool
        """
        return not self == other

    def __str__(self):
        return " + ".join([str(term) for term in self.terms])

    def __len__(self):
        """
        The length of the PauliSum is the number of PauliTerms in the sum.
        """
        return len(self.terms)

    def __getitem__(self, item):
        """
        :param int item: The index of the term in the sum to return
        :return: The PauliTerm at the index-th position in the PauliSum
        :rtype: PauliTerm
        """
        return self.terms[item]

    def __iter__(self):
        return self.terms.__iter__()

    def __mul__(self, other):
        """
        Multiplies together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        :rtype: PauliSum
        """
        if not isinstance(other, (Number, PauliTerm, PauliSum)):
            raise ValueError("Cannot multiply PauliSum by term that is not a Number, PauliTerm, or"
                             "PauliSum")
        elif isinstance(other, PauliSum):
            other_terms = other.terms
        else:
            other_terms = [other]
        new_terms = [lterm * rterm for lterm, rterm in product(self.terms, other_terms)]
        new_sum = PauliSum(new_terms)
        return new_sum.simplify()

    def __rmul__(self, other):
        """
        Multiples together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the multiplication.
        :rtype: PauliSum
        """
        assert isinstance(other, Number)
        new_terms = [term.copy() for term in self.terms]
        for term in new_terms:
            term.coefficient *= other
        return PauliSum(new_terms).simplify()

    def __pow__(self, power):
        """Raises this PauliSum to power.

        :param int power: The power to raise this PauliSum to.
        :return: The power-th power of this PauliSum.
        :rtype: PauliSum
        """
        if not isinstance(power, int) or power < 0:
            raise ValueError("The power must be a non-negative integer.")
        result = PauliSum([ID()])

        if not self.get_qubits():
            # There aren't any nontrivial operators
            terms = [term_with_coeff(term, 1) for term in self.terms]
            for term in terms:
                result *= term
        else:
            for term in self.terms:
                for qubit_id in term.get_qubits():
                    result *= PauliTerm("I", qubit_id)

        for _ in range(power):
            result *= self
        return result

    def __add__(self, other):
        """
        Adds together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the addition.
        :rtype: PauliSum
        """
        if isinstance(other, PauliTerm):
            other = PauliSum([other])
        elif isinstance(other, Number):
            other = PauliSum([other * ID()])
        new_terms = [term.copy() for term in self.terms]
        new_terms.extend(other.terms)
        new_sum = PauliSum(new_terms)
        return new_sum.simplify()

    def __radd__(self, other):
        """
        Adds together this PauliSum with PauliSum, PauliTerm or Number objects. The new term
        is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the addition.
        :rtype: PauliSum
        """
        assert isinstance(other, Number)
        return self + other

    def __sub__(self, other):
        """
        Finds the difference of this PauliSum with PauliSum, PauliTerm or Number objects. The new
        term is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        :rtype: PauliSum
        """
        return self + -1. * other

    def __rsub__(self, other):
        """
        Finds the different of this PauliSum with PauliSum, PauliTerm or Number objects. The new
        term is then simplified according to the Pauli Algebra rules.

        :param other: a PauliSum, PauliTerm or Number object
        :return: A new PauliSum object given by the subtraction.
        :rtype: PauliSum
        """
        return other + -1. * self

    def get_qubits(self):
        """
        The support of all the operators in the PauliSum object.

        :returns: A list of all the qubits in the sum of terms.
        :rtype: list
        """
        return list(set().union(*[term.get_qubits() for term in self.terms]))

    def simplify(self):
        """
        Simplifies the sum of Pauli operators according to Pauli algebra rules.
        """
        return simplify_pauli_sum(self)

    def get_programs(self):
        """
        Get a Pyquil Program corresponding to each term in the PauliSum and a coefficient
        for each program

        :return: (programs, coefficients)
        """
        programs = [term.program for term in self.terms]
        coefficients = np.array([term.coefficient for term in self.terms])
        return programs, coefficients


def simplify_pauli_sum(pauli_sum):
    # You might want to use a defaultdict(list) here, but don't because
    # we want to do our best to preserve the order of terms.
    like_terms = OrderedDict()
    for term in pauli_sum.terms:
        key = term.operations_as_set()
        if key in like_terms:
            like_terms[key].append(term)
        else:
            like_terms[key] = [term]

    terms = []
    for term_list in like_terms.values():
        first_term = term_list[0]
        if len(term_list) == 1 and not np.isclose(first_term.coefficient, 0.0):
            terms.append(first_term)
        else:
            coeff = sum(t.coefficient for t in term_list)
            for t in term_list:
                if list(t._ops.items()) != list(first_term._ops.items()):
                    warnings.warn("The term {} will be combined with {}, but they have different "
                                  "orders of operations. This doesn't matter for QVM or "
                                  "wavefunction simulation but may be important when "
                                  "running on an actual device."
                                  .format(t.id(sort_ops=False), first_term.id(sort_ops=False)))

            if not np.isclose(coeff, 0.0):
                terms.append(term_with_coeff(term_list[0], coeff))
    return PauliSum(terms)


def check_commutation(pauli_list, pauli_two):
    """
    Check if commuting a PauliTerm commutes with a list of other terms by natural calculation.
    Derivation similar to arXiv:1405.5749v2 fo the check_commutation step in
    the Raesi, Wiebe, Sanders algorithm (arXiv:1108.4318, 2011).

    :param list pauli_list: A list of PauliTerm objects
    :param PauliTerm pauli_two_term: A PauliTerm object
    :returns: True if pauli_two object commutes with pauli_list, False otherwise
    :rtype: bool
    """

    def coincident_parity(p1, p2):
        non_similar = 0
        p1_indices = set(p1._ops.keys())
        p2_indices = set(p2._ops.keys())
        for idx in p1_indices.intersection(p2_indices):
            if p1[idx] != p2[idx]:
                non_similar += 1
        return non_similar % 2 == 0

    for i, term in enumerate(pauli_list):
        if not coincident_parity(term, pauli_two):
            return False
    return True


def commuting_sets(pauli_terms):
    """Gather the Pauli terms of pauli_terms variable into commuting sets

    Uses algorithm defined in (Raeisi, Wiebe, Sanders, arXiv:1108.4318, 2011)
    to find commuting sets. Except uses commutation check from arXiv:1405.5749v2

    :param PauliSum pauli_terms: A PauliSum object
    :returns: List of lists where each list contains a commuting set
    :rtype: list
    """

    m_terms = len(pauli_terms.terms)
    m_s = 1
    groups = []
    groups.append([pauli_terms.terms[0]])
    for j in range(1, m_terms):
        isAssigned_bool = False
        for p in range(m_s):  # check if it commutes with each group
            if isAssigned_bool is False:

                if check_commutation(groups[p], pauli_terms.terms[j]):
                    isAssigned_bool = True
                    groups[p].append(pauli_terms.terms[j])
        if isAssigned_bool is False:
            m_s += 1
            groups.append([pauli_terms.terms[j]])
    return groups


def is_identity(term):
    """
    Check if Pauli Term is a scalar multiple of identity

    :param PauliTerm term: A PauliTerm object
    :returns: True if the PauliTerm is a scalar multiple of identity, false otherwise
    :rtype: bool
    """
    return len(term) == 0


def exponentiate(term):
    """
    Creates a pyQuil program that simulates the unitary evolution exp(-1j * term)

    :param PauliTerm term: Tests is a PauliTerm is the identity operator
    :returns: A Program object
    :rtype: Program
    """
    return exponential_map(term)(1.0)


def exponential_map(term):
    """
    Creates map alpha -> exp(-1j*alpha*term) represented as a Program.

    :param PauliTerm term: Tests is a PauliTerm is the identity operator
    :returns: Program
    :rtype: Function
    """
    if not np.isclose(np.imag(term.coefficient), 0.0):
        raise TypeError("PauliTerm coefficient must be real")

    coeff = term.coefficient.real
    term.coefficient = term.coefficient.real

    def exp_wrap(param):
        prog = Program()
        if is_identity(term):
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff, 0))
            prog.inst(X(0))
            prog.inst(PHASE(-param * coeff, 0))
        else:
            prog += _exponentiate_general_case(term, param)
        return prog

    return exp_wrap


def exponentiate_commuting_pauli_sum(pauli_sum):
    """
    Returns a function that maps all substituent PauliTerms and sums them into a program. NOTE: Use
    this function with care. Substituent PauliTerms should commute.

    :param PauliSum pauli_sum: PauliSum to exponentiate.
    :returns: A function that parametrizes the exponential.
    :rtype: function
    """
    if not isinstance(pauli_sum, PauliSum):
        raise TypeError("Argument 'pauli_sum' must be a PauliSum.")

    fns = [exponential_map(term) for term in pauli_sum]

    def combined_exp_wrap(param):
        return Program([f(param) for f in fns])

    return combined_exp_wrap


def _exponentiate_general_case(pauli_term, param):
    """
    Returns a Quil (Program()) object corresponding to the exponential of
    the pauli_term object, i.e. exp[-1.0j * param * pauli_term]

    :param PauliTerm pauli_term: A PauliTerm to exponentiate
    :param float param: scalar, non-complex, value
    :returns: A Quil program object
    :rtype: Program
    """

    def reverse_hack(p):
        # A hack to produce a *temporary* program which reverses p.
        revp = Program()
        revp.inst(list(reversed(p.instructions)))
        return revp

    quil_prog = Program()
    change_to_z_basis = Program()
    change_to_original_basis = Program()
    cnot_seq = Program()
    prev_index = None
    highest_target_index = None

    for index, op in pauli_term:
        if 'X' == op:
            change_to_z_basis.inst(H(index))
            change_to_original_basis.inst(H(index))

        elif 'Y' == op:
            change_to_z_basis.inst(RX(np.pi / 2.0, index))
            change_to_original_basis.inst(RX(-np.pi / 2.0, index))

        elif 'I' == op:
            continue

        if prev_index is not None:
            cnot_seq.inst(CNOT(prev_index, index))

        prev_index = index
        highest_target_index = index

    # building rotation circuit
    quil_prog += change_to_z_basis
    quil_prog += cnot_seq
    quil_prog.inst(RZ(2.0 * pauli_term.coefficient * param, highest_target_index))
    quil_prog += reverse_hack(cnot_seq)
    quil_prog += change_to_original_basis

    return quil_prog


def suzuki_trotter(trotter_order, trotter_steps):
    """
    Generate trotterization coefficients for a given number of Trotter steps.

    U = exp(A + B) is approximated as exp(w1*o1)exp(w2*o2)... This method returns
    a list [(w1, o1), (w2, o2), ... , (wm, om)] of tuples where o=0 corresponds
    to the A operator, o=1 corresponds to the B operator, and w is the
    coefficient in the exponential. For example, a second order Suzuki-Trotter
    approximation to exp(A + B) results in the following
    [(0.5/trotter_steps, 0), (1/trotteri_steps, 1),
    (0.5/trotter_steps, 0)] * trotter_steps.

    :param int trotter_order: order of Suzuki-Trotter approximation
    :param int trotter_steps: number of steps in the approximation
    :returns: List of tuples corresponding to the coefficient and operator
              type: o=0 is A and o=1 is B.
    :rtype: list
    """
    p1 = p2 = p4 = p5 = 1.0 / (4 - (4 ** (1. / 3)))
    p3 = 1 - 4 * p1
    trotter_dict = {1: [(1, 0), (1, 1)],
                    2: [(0.5, 0), (1, 1), (0.5, 0)],
                    3: [(7.0 / 24, 0), (2.0 / 3.0, 1), (3.0 / 4.0, 0), (-2.0 / 3.0, 1),
                        (-1.0 / 24, 0), (1.0, 1)],
                    4: [(p5 / 2, 0), (p5, 1), (p5 / 2, 0),
                        (p4 / 2, 0), (p4, 1), (p4 / 2, 0),
                        (p3 / 2, 0), (p3, 1), (p3 / 2, 0),
                        (p2 / 2, 0), (p2, 1), (p2 / 2, 0),
                        (p1 / 2, 0), (p1, 1), (p1 / 2, 0)]}

    order_slices = [(x0 / trotter_steps, x1) for x0, x1 in trotter_dict[trotter_order]]
    order_slices = order_slices * trotter_steps
    return order_slices


def is_zero(pauli_object):
    """
    Tests to see if a PauliTerm or PauliSum is zero.

    :param pauli_object: Either a PauliTerm or PauliSum
    :returns: True if PauliTerm is zero, False otherwise
    :rtype: bool
    """
    if isinstance(pauli_object, PauliTerm):
        if pauli_object.id() == '':
            return True
        else:
            return False
    elif isinstance(pauli_object, PauliSum):
        if len(pauli_object.terms) == 1 and pauli_object.terms[0].id() == '':
            return True
        else:
            return False
    else:
        raise TypeError("is_zero only checks PauliTerms and PauliSum objects!")


def trotterize(first_pauli_term, second_pauli_term, trotter_order=1,
               trotter_steps=1):
    """
    Create a Quil program that approximates exp( (A + B)t) where A and B are
    PauliTerm operators.

    :param PauliTerm first_pauli_term: PauliTerm denoted `A`
    :param PauliTerm second_pauli_term: PauliTerm denoted `B`
    :param int trotter_order: Optional argument indicating the Suzuki-Trotter
                          approximation order--only accepts orders 1, 2, 3, 4.
    :param int trotter_steps: Optional argument indicating the number of products
                          to decompose the exponential into.

    :return: Quil program
    :rtype: Program
    """

    if not (1 <= trotter_order < 5):
        raise ValueError("trotterize only accepts trotter_order in {1, 2, 3, 4}.")

    commutator = (first_pauli_term * second_pauli_term) + \
                 (-1 * second_pauli_term * first_pauli_term)

    prog = Program()
    if is_zero(commutator):
        param_exp_prog_one = exponential_map(first_pauli_term)
        exp_prog = param_exp_prog_one(1)
        prog += exp_prog
        param_exp_prog_two = exponential_map(second_pauli_term)
        exp_prog = param_exp_prog_two(1)
        prog += exp_prog
        return prog

    order_slices = suzuki_trotter(trotter_order, trotter_steps)
    for coeff, operator in order_slices:
        if operator == 0:
            param_prog = exponential_map(coeff * first_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
        else:
            param_prog = exponential_map(coeff * second_pauli_term)
            exp_prog = param_prog(1)
            prog += exp_prog
    return prog



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and defining parametric programs.
"""
import inspect
from copy import copy
from six.moves import range

from .slot import Slot
from .quil import Program


def argument_count(thing):
    """
    Get the number of arguments a callable has.

    :param thing: A callable.
    :return: The number of arguments it takes.
    :rtype: int
    """
    if not callable(thing):
        raise TypeError("should be callable")
    argspec = inspect.getargspec(thing)
    if argspec.varargs is not None:
        raise RuntimeError("callable should have no *args parameter")
    if argspec.keywords is not None:
        raise RuntimeError("callable should have no **kwargs parameter")
    return len(argspec.args)


class ParametricProgram(object):
    """
    .. note:: Experimental

    A class representing Programs with changeable gate parameters.
    """

    def __init__(self, program_constructor):
        self.num_arguments = argument_count(program_constructor)
        self.slots = [Slot() for _ in range(self.num_arguments)]
        self.instantiated_program = program_constructor(*self.slots)
        if not isinstance(self.instantiated_program, Program):
            raise TypeError("program_constructor should produce a Program object")

    def fuse(self, other):
        """
        .. note:: Experimental

        Fuse another program to this one.

        :param other: A Program or ParametricProgram.
        :return: A new ParametricProgram.
        :rtype: ParametricProgram
        """
        r = copy(self)   # shallow copy the object

        if isinstance(other, ParametricProgram):
            r.num_arguments += other.num_arguments
            r.slots.extend(other.slots)
            r.instantiated_program += other.instantiated_program
        elif isinstance(other, Program):
            r.instantiated_program += other
        else:
            raise TypeError("Can only fuse Programs and ParametricPrograms")

        return r

    def __call__(self, *values):
        if len(values) != self.num_arguments:
            raise RuntimeError("Invalid number of arguments provided"
                               " to a ParametricProgram instance")
        for slot, value in zip(self.slots, values):
            slot._value = value

        return self.instantiated_program


def parametric(decorated_function):
    """
    .. note:: Experimental

    A decorator to change a function into a ParametricProgram.

    :param decorated_function: The function taking parameters producing a Program object.
    :return: a callable ParametricProgram
    :rtype: ParametricProgram
    """
    return ParametricProgram(decorated_function)



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import warnings

from abc import ABC, abstractmethod

from rpcq.messages import ParameterAref

from pyquil.api._error_reporting import _record_call


class QAMError(RuntimeError):
    pass


class QAM(ABC):
    """
    The platonic ideal of this class is as a generic interface describing how a classical computer
    interacts with a live quantum computer.  Eventually, it will turn into a thin layer over the
    QPU and QVM's "QPI" interfaces.

    The reality is that neither the QPU nor the QVM currently support a full-on QPI interface,
    and so the undignified job of this class is to collect enough state that it can convincingly
    pretend to be a QPI-compliant quantum computer.
    """

    @_record_call
    def __init__(self):
        self._variables_shim = {}
        self._n_shots = None
        self._n_bits = None
        self._executable = None
        self._bitstrings = None

        self.status = 'connected'

    @_record_call
    def load(self, executable):
        """
        Initialize a QAM into a fresh state.

        :param executable: Load a compiled executable onto the QAM.
        """
        if self.status == 'loaded':
            warnings.warn("Overwriting previously loaded executable.")
        assert self.status in ['connected', 'done', 'loaded']

        self._variables_shim = {}
        self._executable = executable
        self._bitstrings = None
        self.status = 'loaded'
        return self

    @_record_call
    def write_memory(self, *, region_name: str, offset: int = 0, value=None):
        """
        Writes a value into a memory region on the QAM at a specified offset.

        :param region_name: Name of the declared memory region on the QAM.
        :param offset: Integer offset into the memory region to write to.
        :param value: Value to store at the indicated location.
        """
        assert self.status in ['loaded', 'done']

        aref = ParameterAref(name=region_name, index=offset)
        self._variables_shim[aref] = value

        return self

    @abstractmethod
    def run(self):
        """
        Reset the program counter on a QAM and run its loaded Quil program.
        """
        self.status = 'running'

        return self

    @_record_call
    def wait(self):
        """
        Blocks until the QPU enters the halted state.
        """
        assert self.status == 'running'
        self.status = 'done'
        return self

    @_record_call
    def read_from_memory_region(self, *, region_name: str):
        """
        Reads from a memory region named region_name on the QAM.

        This is a shim over the eventual API and only can return memory from a region named
        "ro" of type ``BIT``.

        :param region_name: The string naming the declared memory region.
        :return: A list of values of the appropriate type.
        """
        assert self.status == 'done'
        if region_name != "ro":
            raise QAMError("Currently only allowed to read measurement data from ro.")
        if self._bitstrings is None:
            raise QAMError("Bitstrings have not yet been populated. Something has gone wrong.")

        return self._bitstrings



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for creating and verifying noisy gate definitions in terms of Kraus maps.
"""
import warnings

# for backwards compatibility
# noinspection PyUnresolvedReferences
from pyquil.noise import (
    append_kraus_to_gate,
    damping_kraus_map,
    dephasing_kraus_map,
    tensor_kraus_maps,
    combine_kraus_maps,
    damping_after_dephasing, add_decoherence_noise)

warnings.warn("pyquil.kraus is deprecated, please use pyquil.noise instead.", DeprecationWarning)


def add_noise_to_program(prog, T1=30e-6, T2=30e-6, gate_time_1q=50e-9, gate_time_2q=150e-09,
                         ro_fidelity=0.95):
    """
    Add generic damping and dephasing noise to a program.

    .. warning::

        This function is deprecated. Please use :py:func:`add_decoherence_noise` instead.

    :param prog: A pyquil program consisting of I, RZ, CZ, and RX(+-pi/2) instructions
    :param Union[Dict[int,float],float] T1: The T1 amplitude damping time either globally or in a
        dictionary indexed by qubit id. By default, this is 30 us.
    :param Union[Dict[int,float],float] T2: The T2 dephasing time either globally or in a
        dictionary indexed by qubit id. By default, this is also 30 us.
    :param float gate_time_1q: The duration of the one-qubit gates, namely RX(+pi/2) and RX(-pi/2).
        By default, this is 50 ns.
    :param float gate_time_2q: The duration of the two-qubit gates, namely CZ.
        By default, this is 150 ns.
    :param Union[Dict[int,float],float] ro_fidelity: The readout assignment fidelity
        :math:`F = (p(0|0) + p(1|1))/2` either globally or in a dictionary indexed by qubit id.
    :return: A new program with noisy operators.
    """
    warnings.warn("pyquil.kraus.add_noise_to_program is deprecated, please use "
                  "pyquil.noise.add_decoherence_noise instead.",
                  DeprecationWarning)
    return add_decoherence_noise(prog, T1=T1, T2=T2, gate_time_1q=gate_time_1q,
                                 gate_time_2q=gate_time_2q, ro_fidelity=ro_fidelity)



# Generated from Quil.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .QuilParser import QuilParser
else:
    from QuilParser import QuilParser

# This class defines a complete listener for a parse tree produced by QuilParser.
class QuilListener(ParseTreeListener):

    # Enter a parse tree produced by QuilParser#quil.
    def enterQuil(self, ctx:QuilParser.QuilContext):
        pass

    # Exit a parse tree produced by QuilParser#quil.
    def exitQuil(self, ctx:QuilParser.QuilContext):
        pass


    # Enter a parse tree produced by QuilParser#allInstr.
    def enterAllInstr(self, ctx:QuilParser.AllInstrContext):
        pass

    # Exit a parse tree produced by QuilParser#allInstr.
    def exitAllInstr(self, ctx:QuilParser.AllInstrContext):
        pass


    # Enter a parse tree produced by QuilParser#instr.
    def enterInstr(self, ctx:QuilParser.InstrContext):
        pass

    # Exit a parse tree produced by QuilParser#instr.
    def exitInstr(self, ctx:QuilParser.InstrContext):
        pass


    # Enter a parse tree produced by QuilParser#gate.
    def enterGate(self, ctx:QuilParser.GateContext):
        pass

    # Exit a parse tree produced by QuilParser#gate.
    def exitGate(self, ctx:QuilParser.GateContext):
        pass


    # Enter a parse tree produced by QuilParser#name.
    def enterName(self, ctx:QuilParser.NameContext):
        pass

    # Exit a parse tree produced by QuilParser#name.
    def exitName(self, ctx:QuilParser.NameContext):
        pass


    # Enter a parse tree produced by QuilParser#qubit.
    def enterQubit(self, ctx:QuilParser.QubitContext):
        pass

    # Exit a parse tree produced by QuilParser#qubit.
    def exitQubit(self, ctx:QuilParser.QubitContext):
        pass


    # Enter a parse tree produced by QuilParser#param.
    def enterParam(self, ctx:QuilParser.ParamContext):
        pass

    # Exit a parse tree produced by QuilParser#param.
    def exitParam(self, ctx:QuilParser.ParamContext):
        pass


    # Enter a parse tree produced by QuilParser#defGate.
    def enterDefGate(self, ctx:QuilParser.DefGateContext):
        pass

    # Exit a parse tree produced by QuilParser#defGate.
    def exitDefGate(self, ctx:QuilParser.DefGateContext):
        pass


    # Enter a parse tree produced by QuilParser#variable.
    def enterVariable(self, ctx:QuilParser.VariableContext):
        pass

    # Exit a parse tree produced by QuilParser#variable.
    def exitVariable(self, ctx:QuilParser.VariableContext):
        pass


    # Enter a parse tree produced by QuilParser#matrix.
    def enterMatrix(self, ctx:QuilParser.MatrixContext):
        pass

    # Exit a parse tree produced by QuilParser#matrix.
    def exitMatrix(self, ctx:QuilParser.MatrixContext):
        pass


    # Enter a parse tree produced by QuilParser#matrixRow.
    def enterMatrixRow(self, ctx:QuilParser.MatrixRowContext):
        pass

    # Exit a parse tree produced by QuilParser#matrixRow.
    def exitMatrixRow(self, ctx:QuilParser.MatrixRowContext):
        pass


    # Enter a parse tree produced by QuilParser#defCircuit.
    def enterDefCircuit(self, ctx:QuilParser.DefCircuitContext):
        pass

    # Exit a parse tree produced by QuilParser#defCircuit.
    def exitDefCircuit(self, ctx:QuilParser.DefCircuitContext):
        pass


    # Enter a parse tree produced by QuilParser#qubitVariable.
    def enterQubitVariable(self, ctx:QuilParser.QubitVariableContext):
        pass

    # Exit a parse tree produced by QuilParser#qubitVariable.
    def exitQubitVariable(self, ctx:QuilParser.QubitVariableContext):
        pass


    # Enter a parse tree produced by QuilParser#circuitQubit.
    def enterCircuitQubit(self, ctx:QuilParser.CircuitQubitContext):
        pass

    # Exit a parse tree produced by QuilParser#circuitQubit.
    def exitCircuitQubit(self, ctx:QuilParser.CircuitQubitContext):
        pass


    # Enter a parse tree produced by QuilParser#circuitGate.
    def enterCircuitGate(self, ctx:QuilParser.CircuitGateContext):
        pass

    # Exit a parse tree produced by QuilParser#circuitGate.
    def exitCircuitGate(self, ctx:QuilParser.CircuitGateContext):
        pass


    # Enter a parse tree produced by QuilParser#circuitInstr.
    def enterCircuitInstr(self, ctx:QuilParser.CircuitInstrContext):
        pass

    # Exit a parse tree produced by QuilParser#circuitInstr.
    def exitCircuitInstr(self, ctx:QuilParser.CircuitInstrContext):
        pass


    # Enter a parse tree produced by QuilParser#circuit.
    def enterCircuit(self, ctx:QuilParser.CircuitContext):
        pass

    # Exit a parse tree produced by QuilParser#circuit.
    def exitCircuit(self, ctx:QuilParser.CircuitContext):
        pass


    # Enter a parse tree produced by QuilParser#measure.
    def enterMeasure(self, ctx:QuilParser.MeasureContext):
        pass

    # Exit a parse tree produced by QuilParser#measure.
    def exitMeasure(self, ctx:QuilParser.MeasureContext):
        pass


    # Enter a parse tree produced by QuilParser#addr.
    def enterAddr(self, ctx:QuilParser.AddrContext):
        pass

    # Exit a parse tree produced by QuilParser#addr.
    def exitAddr(self, ctx:QuilParser.AddrContext):
        pass


    # Enter a parse tree produced by QuilParser#defLabel.
    def enterDefLabel(self, ctx:QuilParser.DefLabelContext):
        pass

    # Exit a parse tree produced by QuilParser#defLabel.
    def exitDefLabel(self, ctx:QuilParser.DefLabelContext):
        pass


    # Enter a parse tree produced by QuilParser#label.
    def enterLabel(self, ctx:QuilParser.LabelContext):
        pass

    # Exit a parse tree produced by QuilParser#label.
    def exitLabel(self, ctx:QuilParser.LabelContext):
        pass


    # Enter a parse tree produced by QuilParser#halt.
    def enterHalt(self, ctx:QuilParser.HaltContext):
        pass

    # Exit a parse tree produced by QuilParser#halt.
    def exitHalt(self, ctx:QuilParser.HaltContext):
        pass


    # Enter a parse tree produced by QuilParser#jump.
    def enterJump(self, ctx:QuilParser.JumpContext):
        pass

    # Exit a parse tree produced by QuilParser#jump.
    def exitJump(self, ctx:QuilParser.JumpContext):
        pass


    # Enter a parse tree produced by QuilParser#jumpWhen.
    def enterJumpWhen(self, ctx:QuilParser.JumpWhenContext):
        pass

    # Exit a parse tree produced by QuilParser#jumpWhen.
    def exitJumpWhen(self, ctx:QuilParser.JumpWhenContext):
        pass


    # Enter a parse tree produced by QuilParser#jumpUnless.
    def enterJumpUnless(self, ctx:QuilParser.JumpUnlessContext):
        pass

    # Exit a parse tree produced by QuilParser#jumpUnless.
    def exitJumpUnless(self, ctx:QuilParser.JumpUnlessContext):
        pass


    # Enter a parse tree produced by QuilParser#resetState.
    def enterResetState(self, ctx:QuilParser.ResetStateContext):
        pass

    # Exit a parse tree produced by QuilParser#resetState.
    def exitResetState(self, ctx:QuilParser.ResetStateContext):
        pass


    # Enter a parse tree produced by QuilParser#wait.
    def enterWait(self, ctx:QuilParser.WaitContext):
        pass

    # Exit a parse tree produced by QuilParser#wait.
    def exitWait(self, ctx:QuilParser.WaitContext):
        pass


    # Enter a parse tree produced by QuilParser#memoryDescriptor.
    def enterMemoryDescriptor(self, ctx:QuilParser.MemoryDescriptorContext):
        pass

    # Exit a parse tree produced by QuilParser#memoryDescriptor.
    def exitMemoryDescriptor(self, ctx:QuilParser.MemoryDescriptorContext):
        pass


    # Enter a parse tree produced by QuilParser#offsetDescriptor.
    def enterOffsetDescriptor(self, ctx:QuilParser.OffsetDescriptorContext):
        pass

    # Exit a parse tree produced by QuilParser#offsetDescriptor.
    def exitOffsetDescriptor(self, ctx:QuilParser.OffsetDescriptorContext):
        pass


    # Enter a parse tree produced by QuilParser#classicalUnary.
    def enterClassicalUnary(self, ctx:QuilParser.ClassicalUnaryContext):
        pass

    # Exit a parse tree produced by QuilParser#classicalUnary.
    def exitClassicalUnary(self, ctx:QuilParser.ClassicalUnaryContext):
        pass


    # Enter a parse tree produced by QuilParser#classicalBinary.
    def enterClassicalBinary(self, ctx:QuilParser.ClassicalBinaryContext):
        pass

    # Exit a parse tree produced by QuilParser#classicalBinary.
    def exitClassicalBinary(self, ctx:QuilParser.ClassicalBinaryContext):
        pass


    # Enter a parse tree produced by QuilParser#logicalBinaryOp.
    def enterLogicalBinaryOp(self, ctx:QuilParser.LogicalBinaryOpContext):
        pass

    # Exit a parse tree produced by QuilParser#logicalBinaryOp.
    def exitLogicalBinaryOp(self, ctx:QuilParser.LogicalBinaryOpContext):
        pass


    # Enter a parse tree produced by QuilParser#arithmeticBinaryOp.
    def enterArithmeticBinaryOp(self, ctx:QuilParser.ArithmeticBinaryOpContext):
        pass

    # Exit a parse tree produced by QuilParser#arithmeticBinaryOp.
    def exitArithmeticBinaryOp(self, ctx:QuilParser.ArithmeticBinaryOpContext):
        pass


    # Enter a parse tree produced by QuilParser#move.
    def enterMove(self, ctx:QuilParser.MoveContext):
        pass

    # Exit a parse tree produced by QuilParser#move.
    def exitMove(self, ctx:QuilParser.MoveContext):
        pass


    # Enter a parse tree produced by QuilParser#exchange.
    def enterExchange(self, ctx:QuilParser.ExchangeContext):
        pass

    # Exit a parse tree produced by QuilParser#exchange.
    def exitExchange(self, ctx:QuilParser.ExchangeContext):
        pass


    # Enter a parse tree produced by QuilParser#convert.
    def enterConvert(self, ctx:QuilParser.ConvertContext):
        pass

    # Exit a parse tree produced by QuilParser#convert.
    def exitConvert(self, ctx:QuilParser.ConvertContext):
        pass


    # Enter a parse tree produced by QuilParser#load.
    def enterLoad(self, ctx:QuilParser.LoadContext):
        pass

    # Exit a parse tree produced by QuilParser#load.
    def exitLoad(self, ctx:QuilParser.LoadContext):
        pass


    # Enter a parse tree produced by QuilParser#store.
    def enterStore(self, ctx:QuilParser.StoreContext):
        pass

    # Exit a parse tree produced by QuilParser#store.
    def exitStore(self, ctx:QuilParser.StoreContext):
        pass


    # Enter a parse tree produced by QuilParser#classicalComparison.
    def enterClassicalComparison(self, ctx:QuilParser.ClassicalComparisonContext):
        pass

    # Exit a parse tree produced by QuilParser#classicalComparison.
    def exitClassicalComparison(self, ctx:QuilParser.ClassicalComparisonContext):
        pass


    # Enter a parse tree produced by QuilParser#nop.
    def enterNop(self, ctx:QuilParser.NopContext):
        pass

    # Exit a parse tree produced by QuilParser#nop.
    def exitNop(self, ctx:QuilParser.NopContext):
        pass


    # Enter a parse tree produced by QuilParser#include.
    def enterInclude(self, ctx:QuilParser.IncludeContext):
        pass

    # Exit a parse tree produced by QuilParser#include.
    def exitInclude(self, ctx:QuilParser.IncludeContext):
        pass


    # Enter a parse tree produced by QuilParser#pragma.
    def enterPragma(self, ctx:QuilParser.PragmaContext):
        pass

    # Exit a parse tree produced by QuilParser#pragma.
    def exitPragma(self, ctx:QuilParser.PragmaContext):
        pass


    # Enter a parse tree produced by QuilParser#pragma_name.
    def enterPragma_name(self, ctx:QuilParser.Pragma_nameContext):
        pass

    # Exit a parse tree produced by QuilParser#pragma_name.
    def exitPragma_name(self, ctx:QuilParser.Pragma_nameContext):
        pass


    # Enter a parse tree produced by QuilParser#numberExp.
    def enterNumberExp(self, ctx:QuilParser.NumberExpContext):
        pass

    # Exit a parse tree produced by QuilParser#numberExp.
    def exitNumberExp(self, ctx:QuilParser.NumberExpContext):
        pass


    # Enter a parse tree produced by QuilParser#powerExp.
    def enterPowerExp(self, ctx:QuilParser.PowerExpContext):
        pass

    # Exit a parse tree produced by QuilParser#powerExp.
    def exitPowerExp(self, ctx:QuilParser.PowerExpContext):
        pass


    # Enter a parse tree produced by QuilParser#mulDivExp.
    def enterMulDivExp(self, ctx:QuilParser.MulDivExpContext):
        pass

    # Exit a parse tree produced by QuilParser#mulDivExp.
    def exitMulDivExp(self, ctx:QuilParser.MulDivExpContext):
        pass


    # Enter a parse tree produced by QuilParser#parenthesisExp.
    def enterParenthesisExp(self, ctx:QuilParser.ParenthesisExpContext):
        pass

    # Exit a parse tree produced by QuilParser#parenthesisExp.
    def exitParenthesisExp(self, ctx:QuilParser.ParenthesisExpContext):
        pass


    # Enter a parse tree produced by QuilParser#variableExp.
    def enterVariableExp(self, ctx:QuilParser.VariableExpContext):
        pass

    # Exit a parse tree produced by QuilParser#variableExp.
    def exitVariableExp(self, ctx:QuilParser.VariableExpContext):
        pass


    # Enter a parse tree produced by QuilParser#addrExp.
    def enterAddrExp(self, ctx:QuilParser.AddrExpContext):
        pass

    # Exit a parse tree produced by QuilParser#addrExp.
    def exitAddrExp(self, ctx:QuilParser.AddrExpContext):
        pass


    # Enter a parse tree produced by QuilParser#signedExp.
    def enterSignedExp(self, ctx:QuilParser.SignedExpContext):
        pass

    # Exit a parse tree produced by QuilParser#signedExp.
    def exitSignedExp(self, ctx:QuilParser.SignedExpContext):
        pass


    # Enter a parse tree produced by QuilParser#addSubExp.
    def enterAddSubExp(self, ctx:QuilParser.AddSubExpContext):
        pass

    # Exit a parse tree produced by QuilParser#addSubExp.
    def exitAddSubExp(self, ctx:QuilParser.AddSubExpContext):
        pass


    # Enter a parse tree produced by QuilParser#functionExp.
    def enterFunctionExp(self, ctx:QuilParser.FunctionExpContext):
        pass

    # Exit a parse tree produced by QuilParser#functionExp.
    def exitFunctionExp(self, ctx:QuilParser.FunctionExpContext):
        pass


    # Enter a parse tree produced by QuilParser#function.
    def enterFunction(self, ctx:QuilParser.FunctionContext):
        pass

    # Exit a parse tree produced by QuilParser#function.
    def exitFunction(self, ctx:QuilParser.FunctionContext):
        pass


    # Enter a parse tree produced by QuilParser#sign.
    def enterSign(self, ctx:QuilParser.SignContext):
        pass

    # Exit a parse tree produced by QuilParser#sign.
    def exitSign(self, ctx:QuilParser.SignContext):
        pass


    # Enter a parse tree produced by QuilParser#number.
    def enterNumber(self, ctx:QuilParser.NumberContext):
        pass

    # Exit a parse tree produced by QuilParser#number.
    def exitNumber(self, ctx:QuilParser.NumberContext):
        pass


    # Enter a parse tree produced by QuilParser#imaginaryN.
    def enterImaginaryN(self, ctx:QuilParser.ImaginaryNContext):
        pass

    # Exit a parse tree produced by QuilParser#imaginaryN.
    def exitImaginaryN(self, ctx:QuilParser.ImaginaryNContext):
        pass


    # Enter a parse tree produced by QuilParser#realN.
    def enterRealN(self, ctx:QuilParser.RealNContext):
        pass

    # Exit a parse tree produced by QuilParser#realN.
    def exitRealN(self, ctx:QuilParser.RealNContext):
        pass





##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for facilitating connections to the QVM / QPU.
"""
import warnings

__all__ = ['QVMConnection', 'LocalQVMCompiler', 'QVMCompiler', 'QPUCompiler',
           'Job', 'Device', 'ForestConnection', 'pyquil_protect',
           'WavefunctionSimulator', 'QuantumComputer', 'list_quantum_computers', 'get_qc',
           'QAM', 'QVM', 'QPU',
           'BenchmarkConnection', 'LocalBenchmarkConnection', 'get_benchmarker']

from pyquil.api._base_connection import ForestConnection
from pyquil.api._benchmark import BenchmarkConnection, LocalBenchmarkConnection, get_benchmarker
from pyquil.api._compiler import QVMCompiler, QPUCompiler, LocalQVMCompiler
from pyquil.api._error_reporting import pyquil_protect
from pyquil.api._job import Job
from pyquil.api._qam import QAM
from pyquil.api._qpu import QPU
from pyquil.api._quantum_computer import QuantumComputer, list_quantum_computers, get_qc, local_qvm
from pyquil.api._qvm import QVMConnection, QVM
from pyquil.api._wavefunction_simulator import WavefunctionSimulator
from pyquil.device import Device


class SyncConnection(QVMConnection):
    def __init__(self, *args, **kwargs):
        warnings.warn("SyncConnection has been renamed to QVMConnection and will be removed in the future",
                      stacklevel=2)
        super(SyncConnection, self).__init__(*args, **kwargs)


class JobConnection(object):
    def __init__(self, *args, **kwargs):
        raise DeprecationWarning("""
JobConnection has been deprecated and will be removed in a future version.
Use QVMConnection instead.

Note that QVMConnection behaves differently than JobConnection did:
run(), run_and_measure(), wavefunction(), and expectation() all now directly
return the result of the program instead of returning a JobResult object.

This means you need to replace constructs like this:
    qvm = JobConnection()
    job = qvm.run(program, ...)
    wait_for_job(job)
    result = job.result()
with just this:
    qvm = JobConnection()
    result = qvm.run(program, ...)

For more information see https://go.rigetti.com/connections\n""")



#!/usr/bin/env python
##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import os
import re
import sys

from setuptools import setup, find_packages

HERE = os.path.abspath(os.path.dirname(__file__))

if sys.version_info < (3, 5):
    raise ImportError('\n'.join([
        'PyQuil 2.0+ requires Python 3'
        '',
        'To install the most recent version with support for Python 2, make sure you',
        'have pip >= 9.0 as well as setuptools >= 24.2:',
        '',
        ' $ pip install pip setuptools --upgrade',
        '',
        'Then you can either',
        '',
        '- install an older version of PyQuil:',
        '',
        " $ pip install 'pyquil<2.0'",
        '',
        '- Upgrade your system to use Python 3.', ]))


def read(*parts):
    with open(os.path.join(HERE, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pyquil",
    version=find_version('pyquil', '__init__.py'),
    author="Rigetti Computing",
    author_email="softapps@rigetti.com",
    description="A Python library to generate Quantum Instruction Language (Quil) Programs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/rigetticomputing/pyquil.git",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    license='Apache-2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        # The minimum spec for a working pyquil install.
        # note to developers: this should be a subset of requirements.txt
        'numpy',
        'antlr4-python3-runtime',
        'requests',
        'contextvars',
        'six',
        'networkx',
        'dataclasses',
        'rpcq==2.0.0'
    ],
    keywords='quantum quil programming hybrid',
    python_requires=">=3.5",
)



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
from __future__ import print_function

import re
import warnings
from json.decoder import JSONDecodeError
from typing import Dict, Union, Sequence

import numpy as np
import requests
from requests.adapters import HTTPAdapter
from six import integer_types
from urllib3 import Retry

from pyquil import Program
from pyquil.api._config import PyquilConfig
from pyquil.api._error_reporting import _record_call
from pyquil.api._errors import error_mapping, UnknownApiError, TooManyQubitsError
from pyquil.device import Specs, ISA
from pyquil.wavefunction import Wavefunction

TYPE_EXPECTATION = "expectation"
TYPE_MULTISHOT = "multishot"
TYPE_MULTISHOT_MEASURE = "multishot-measure"
TYPE_WAVEFUNCTION = "wavefunction"


def get_json(session, url, params: dict = None):
    """
    Get JSON from a Forest endpoint.
    """
    res = session.get(url, params=params)
    if res.status_code >= 400:
        raise parse_error(res)
    return res.json()


def post_json(session, url, json):
    """
    Post JSON to the Forest endpoint.
    """
    res = session.post(url, json=json)
    if res.status_code >= 400:
        raise parse_error(res)
    return res


def parse_error(res):
    """
    Every server error should contain a "status" field with a human readable explanation of
    what went wrong as well as a "error_type" field indicating the kind of error that can be mapped
    to a Python type.

    There's a fallback error UnknownError for other types of exceptions (network issues, api
    gateway problems, etc.)
    """
    try:
        body = res.json()
    except JSONDecodeError:
        raise UnknownApiError(res.text)

    if 'error_type' not in body:
        raise UnknownApiError(str(body))

    error_type = body['error_type']
    status = body['status']

    if re.search(r"[0-9]+ qubits were requested, but the QVM is limited to [0-9]+ qubits.", status):
        return TooManyQubitsError(status)

    error_cls = error_mapping.get(error_type, UnknownApiError)
    return error_cls(status)


def get_session():
    """
    Create a requests session to access the REST API

    :return: requests session
    :rtype: Session
    """
    config = PyquilConfig()
    session = requests.Session()
    retry_adapter = HTTPAdapter(max_retries=Retry(total=3,
                                                  method_whitelist=['POST'],
                                                  status_forcelist=[502, 503, 504, 521, 523],
                                                  backoff_factor=0.2,
                                                  raise_on_status=False))

    session.mount("http://", retry_adapter)
    session.mount("https://", retry_adapter)

    # We need this to get binary payload for the wavefunction call.
    session.headers.update({"Accept": "application/octet-stream",
                            "X-User-Id": config.user_id,
                            "X-Api-Key": config.api_key})

    session.headers.update({
        'Content-Type': 'application/json; charset=utf-8'
    })

    return session


def validate_noise_probabilities(noise_parameter):
    """
    Is noise_parameter a valid specification of noise probabilities for depolarizing noise?

    :param list noise_parameter: List of noise parameter values to be validated.
    """
    if not noise_parameter:
        return
    if not isinstance(noise_parameter, list):
        raise TypeError("noise_parameter must be a list")
    if any([not isinstance(value, float) for value in noise_parameter]):
        raise TypeError("noise_parameter values should all be floats")
    if len(noise_parameter) != 3:
        raise ValueError("noise_parameter lists must be of length 3")
    if sum(noise_parameter) > 1 or sum(noise_parameter) < 0:
        raise ValueError("sum of entries in noise_parameter must be between 0 and 1 (inclusive)")
    if any([value < 0 for value in noise_parameter]):
        raise ValueError("noise_parameter values should all be non-negative")


def validate_qubit_list(qubit_list):
    """
    Check the validity of qubits for the payload.

    :param list|range qubit_list: List of qubits to be validated.
    """
    if not isinstance(qubit_list, (list, range)):
        raise TypeError("run_items must be a list")
    if any(not isinstance(i, integer_types) or i < 0 for i in qubit_list):
        raise TypeError("run_items list must contain positive integer values")
    return qubit_list


def prepare_register_list(register_dict: Dict[str, Union[bool, Sequence[int]]]):
    """
    Canonicalize classical addresses for the payload and ready MemoryReference instances
    for serialization.

    This function will cast keys that are iterables of int-likes to a list of Python
    ints. This is to support specifying the register offsets as ``range()`` or numpy
    arrays. This mutates ``register_dict``.

    :param register_dict: The classical memory to retrieve. Specified as a dictionary:
        the keys are the names of memory regions, and the values are either (1) a list of
        integers for reading out specific entries in that memory region, or (2) True, for
        reading out the entire memory region.
    """
    if not isinstance(register_dict, dict):
        raise TypeError("register_dict must be a dict but got " + repr(register_dict))

    for k, v in register_dict.items():
        if isinstance(v, bool):
            assert v    # If boolean v must be True
            continue

        indices = [int(x) for x in v]  # support ranges, numpy, ...

        if not all(x >= 0 for x in indices):
            raise TypeError("Negative indices into classical arrays are not allowed.")
        register_dict[k] = indices

    return register_dict


def run_and_measure_payload(quil_program, qubits, trials, random_seed):
    """REST payload for :py:func:`ForestConnection._run_and_measure`"""
    if not quil_program:
        raise ValueError("You have attempted to run an empty program."
                         " Please provide gates or measure instructions to your program.")

    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")
    qubits = validate_qubit_list(qubits)
    if not isinstance(trials, integer_types):
        raise TypeError("trials must be an integer")

    payload = {"type": TYPE_MULTISHOT_MEASURE,
               "qubits": list(qubits),
               "trials": trials,
               "compiled-quil": quil_program.out()}

    if random_seed is not None:
        payload['rng-seed'] = random_seed

    return payload


def wavefunction_payload(quil_program, random_seed):
    """REST payload for :py:func:`ForestConnection._wavefunction`"""
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")

    payload = {'type': TYPE_WAVEFUNCTION,
               'compiled-quil': quil_program.out()}

    if random_seed is not None:
        payload['rng-seed'] = random_seed

    return payload


def expectation_payload(prep_prog, operator_programs, random_seed):
    """REST payload for :py:func:`ForestConnection._expectation`"""
    if operator_programs is None:
        operator_programs = [Program()]

    if not isinstance(prep_prog, Program):
        raise TypeError("prep_prog variable must be a Quil program object")

    payload = {'type': TYPE_EXPECTATION,
               'state-preparation': prep_prog.out(),
               'operators': [x.out() for x in operator_programs]}

    if random_seed is not None:
        payload['rng-seed'] = random_seed

    return payload


def qvm_run_payload(quil_program, classical_addresses, trials,
                    measurement_noise, gate_noise, random_seed):
    """REST payload for :py:func:`ForestConnection._qvm_run`"""
    if not quil_program:
        raise ValueError("You have attempted to run an empty program."
                         " Please provide gates or measure instructions to your program.")
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Quil program object")
    classical_addresses = prepare_register_list(classical_addresses)
    if not isinstance(trials, integer_types):
        raise TypeError("trials must be an integer")

    payload = {"type": TYPE_MULTISHOT,
               "addresses": classical_addresses,
               "trials": trials,
               "compiled-quil": quil_program.out()}

    if measurement_noise is not None:
        payload["measurement-noise"] = measurement_noise
    if gate_noise is not None:
        payload["gate-noise"] = gate_noise
    if random_seed is not None:
        payload['rng-seed'] = random_seed

    return payload


def quilc_compile_payload(quil_program, isa, specs):
    """REST payload for :py:func:`ForestConnection._quilc_compile`"""
    if not quil_program:
        raise ValueError("You have attempted to compile an empty program."
                         " Please provide an actual program.")
    if not isinstance(quil_program, Program):
        raise TypeError("quil_program must be a Program object.")
    if not isinstance(isa, ISA):
        raise TypeError("isa must be an ISA object.")
    if not isinstance(specs, Specs):
        raise TypeError("specs must be a Specs object.")

    payload = {"uncompiled-quil": quil_program.out(),
               "target-device": {
                   "isa": isa.to_dict(),
                   "specs": specs.to_dict()}}

    return payload


class ForestConnection:
    @_record_call
    def __init__(self, sync_endpoint=None, compiler_endpoint=None):
        """
        Represents a connection to Forest containing methods to wrap all possible API endpoints.

        Users should not use methods from this class directly.

        :param sync_endpoint: The endpoint of the server for running (small) QVM jobs
        :param compiler_endpoint: The endpoint of the server for running (small) compiler jobs
        """

        if not sync_endpoint:
            pyquil_config = PyquilConfig()
            sync_endpoint = pyquil_config.qvm_url
        if not compiler_endpoint:
            pyquil_config = PyquilConfig()
            compiler_endpoint = pyquil_config.compiler_url

        self.sync_endpoint = sync_endpoint
        self.compiler_endpoint = compiler_endpoint
        self.session = get_session()

    @_record_call
    def _run_and_measure(self, quil_program, qubits, trials, random_seed) -> np.ndarray:
        """
        Run a Forest ``run_and_measure`` job.

        Users should use :py:func:`WavefunctionSimulator.run_and_measure` instead of calling
        this directly.
        """
        payload = run_and_measure_payload(quil_program, qubits, trials, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return np.asarray(response.json())

    @_record_call
    def _wavefunction(self, quil_program, random_seed) -> Wavefunction:
        """
        Run a Forest ``wavefunction`` job.

        Users should use :py:func:`WavefunctionSimulator.wavefunction` instead of calling
        this directly.
        """

        payload = wavefunction_payload(quil_program, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return Wavefunction.from_bit_packed_string(response.content)

    @_record_call
    def _expectation(self, prep_prog, operator_programs, random_seed) -> np.ndarray:
        """
        Run a Forest ``expectation`` job.

        Users should use :py:func:`WavefunctionSimulator.expectation` instead of calling
        this directly.
        """
        if isinstance(operator_programs, Program):
            warnings.warn("You have provided a Program rather than a list of Programs. The results "
                          "from expectation will be line-wise expectation values of the "
                          "operator_programs.", SyntaxWarning)

        payload = expectation_payload(prep_prog, operator_programs, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)
        return np.asarray(response.json())

    @_record_call
    def _qvm_run(self, quil_program, classical_addresses, trials,
                 measurement_noise, gate_noise, random_seed) -> np.ndarray:
        """
        Run a Forest ``run`` job on a QVM.

        Users should use :py:func:`QVM.run` instead of calling this directly.
        """
        payload = qvm_run_payload(quil_program, classical_addresses, trials,
                                  measurement_noise, gate_noise, random_seed)
        response = post_json(self.session, self.sync_endpoint + "/qvm", payload)

        ram = response.json()

        for k in ram.keys():
            ram[k] = np.array(ram[k])

        return ram

    @_record_call
    def _qvm_get_version_info(self) -> dict:
        """
        Return version information for the QVM.

        :return: Dictionary with version information
        """
        response = post_json(self.session, self.sync_endpoint, {'type': 'version'})
        split_version_string = response.text.split()
        try:
            qvm_app_version = split_version_string[0]
            qvm_lib_version = split_version_string[2][:-1]
        except ValueError:
            raise TypeError(f'Malformed version string returned by the QVM: {response.text}')
        return {'qvm-app': qvm_app_version, 'qvm-lib': qvm_lib_version}

    def _quilc_compile(self, quil_program, isa, specs):
        """
        Sends a quilc job to Forest.

        Users should use :py:func:`LocalCompiler.quil_to_native_quil` instead of calling this
        directly.
        """
        payload = quilc_compile_payload(quil_program, isa, specs)
        response = post_json(self.session, self.sync_endpoint + "/quilc", payload)
        unpacked_response = response.json()
        return unpacked_response

    def _quilc_get_version_info(self) -> dict:
        """
        Return version information for quilc.

        :return: Dictionary with version information
        """
        return get_json(self.session, self.sync_endpoint + '/version')



##############################################################################
# Copyright 2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import rpcq
from rpcq import Client
from rpcq.messages import (RandomizedBenchmarkingRequest, RandomizedBenchmarkingResponse,
                           ConjugateByCliffordRequest, ConjugateByCliffordResponse)

from pyquil.api._base_connection import get_session, post_json
from pyquil.api._config import PyquilConfig
from pyquil.api._error_reporting import _record_call
from pyquil.api._qac import AbstractBenchmarker
from pyquil.paulis import PauliTerm
from pyquil.quil import address_qubits, Program


class BenchmarkConnection(AbstractBenchmarker):
    """
    Represents a connection to a server that generates benchmarking data.
    """

    @_record_call
    def __init__(self, endpoint=None):
        """
        Client to communicate with the benchmarking data endpoint.

        :param endpoint: TCP or IPC endpoint of the Compiler Server
        """

        self.client = Client(endpoint)

    @_record_call
    def apply_clifford_to_pauli(self, clifford, pauli_in):
        r"""
        Given a circuit that consists only of elements of the Clifford group,
        return its action on a PauliTerm.

        In particular, for Clifford C, and Pauli P, this returns the PauliTerm
        representing PCP^{\dagger}.

        :param Program clifford: A Program that consists only of Clifford operations.
        :param PauliTerm pauli_in: A PauliTerm to be acted on by clifford via conjugation.
        :return: A PauliTerm corresponding to pauli_in * clifford * pauli_in^{\dagger}
        """

        indices_and_terms = list(zip(*list(pauli_in.operations_as_set())))

        payload = ConjugateByCliffordRequest(
            clifford=clifford.out(),
            pauli=rpcq.messages.PauliTerm(
                indices=list(indices_and_terms[0]), symbols=list(indices_and_terms[1])))
        response: ConjugateByCliffordResponse = self.client.call(
            'conjugate_pauli_by_clifford', payload)
        phase_factor, paulis = response.phase, response.pauli

        pauli_out = PauliTerm("I", 0, 1.j ** phase_factor)
        clifford_qubits = clifford.get_qubits()
        pauli_qubits = pauli_in.get_qubits()
        all_qubits = sorted(set(pauli_qubits).union(set(clifford_qubits)))
        # The returned pauli will have specified its value on all_qubits, sorted by index.
        #  This is maximal set of qubits that can be affected by this conjugation.
        for i, pauli in enumerate(paulis):
            pauli_out *= PauliTerm(pauli, all_qubits[i])
        return pauli_out * pauli_in.coefficient

    @_record_call
    def generate_rb_sequence(self, depth, gateset, seed=None):
        """
        Construct a randomized benchmarking experiment on the given qubits, decomposing into
        gateset.

        The JSON payload that is parsed is a list of lists of indices, or Nones. In the
        former case, they are the index of the gate in the gateset.

        :param int depth: The number of Clifford gates to include in the randomized benchmarking
         experiment. This is different than the number of gates in the resulting experiment.
        :param list gateset: A list of pyquil gates to decompose the Clifford elements into. These
         must generate the clifford group on the qubits of interest. e.g. for one qubit
         [RZ(np.pi/2), RX(np.pi/2)].
        :param int seed: A positive integer that seeds the random generation of the gate sequence.
        :return: A list of pyquil programs. Each pyquil program is a circuit that represents an
         element of the Clifford group. When these programs are composed, the resulting Program
         will be the randomized benchmarking experiment of the desired depth. e.g. if the return
         programs are called cliffords then `sum(cliffords, Program())` will give the randomized
         benchmarking experiment, which will compose to the identity program.
        """

        # Support QubitPlaceholders: we temporarily index to arbitrary integers.
        # `generate_rb_sequence` handles mapping back to the original gateset gates.
        gateset_as_program = address_qubits(sum(gateset, Program()))
        qubits = len(gateset_as_program.get_qubits())
        gateset_for_api = gateset_as_program.out().splitlines()

        depth = int(depth)  # needs to be jsonable, no np.int64 please!

        payload = RandomizedBenchmarkingRequest(depth=depth, qubits=qubits, gateset=gateset_for_api, seed=seed)
        response = self.client.call('generate_rb_sequence', payload)  # type: RandomizedBenchmarkingResponse

        programs = []
        for clifford in response.sequence:
            clifford_program = Program()
            # Like below, we reversed the order because the API currently hands back the Clifford
            # decomposition right-to-left.
            for index in reversed(clifford):
                clifford_program.inst(gateset[index])
            programs.append(clifford_program)
        # The programs are returned in "textbook style" right-to-left order. To compose them into
        #  the correct pyquil program, we reverse the order.
        return list(reversed(programs))


class LocalBenchmarkConnection(AbstractBenchmarker):
    """
    Represents a connection to a locally-running server that generates randomized benchmarking data.
    """

    @_record_call
    def __init__(self, endpoint=None):
        super(LocalBenchmarkConnection, self).__init__()
        self.endpoint = endpoint
        self.session = get_session()

    @staticmethod
    def _clifford_application_payload(clifford, pauli):
        """
        Prepares a JSON payload for conjugating a Pauli by a Clifford.

         See :py:func:`apply_clifford_to_pauli`.

        :param Program clifford: A Program that consists only of Clifford operations.
        :param PauliTerm pauli: A PauliTerm to be acted on by clifford via conjugation.
        :return: The JSON payload, with keys "clifford" and "pauli".
        """
        indices_and_terms = zip(*list(pauli.operations_as_set()))
        return {"clifford": clifford.out(),
                "pauli": list(indices_and_terms)}

    @_record_call
    def apply_clifford_to_pauli(self, clifford, pauli_in):
        r"""
        Given a circuit that consists only of elements of the Clifford group,
        return its action on a PauliTerm.

        In particular, for Clifford C, and Pauli P, this returns the PauliTerm
        representing PCP^{\dagger}.

        :param Program clifford: A Program that consists only of Clifford operations.
        :param PauliTerm pauli_in: A PauliTerm to be acted on by clifford via conjugation.
        :return: A PauliTerm corresponding to pauli_in * clifford * pauli_in^{\dagger}
        """
        payload = self._clifford_application_payload(clifford, pauli_in)
        phase_factor, paulis = post_json(self.session, self.endpoint + "/apply-clifford",
                                         payload).json()
        pauli_out = PauliTerm("I", 0, 1.j ** phase_factor)
        clifford_qubits = clifford.get_qubits()
        pauli_qubits = pauli_in.get_qubits()
        all_qubits = sorted(set(pauli_qubits).union(set(clifford_qubits)))
        # The returned pauli will have specified its value on all_qubits, sorted by index.
        #  This is maximal set of qubits that can be affected by this conjugation.
        for i, pauli in enumerate(paulis):
            pauli_out *= PauliTerm(pauli, all_qubits[i])
        return pauli_out * pauli_in.coefficient

    @staticmethod
    def _rb_sequence_payload(depth, gateset, seed=None):
        """
        Prepares a JSON payload for generating a randomized benchmarking sequence.

        See :py:func:`generate_rb_sequence`.

        :param int depth: The number of cliffords per rb sequences to generate.
        :param list gateset: A list of Gate objects that make up the gateset to decompose
            the Cliffords into.
        :return: The JSON payload, with keys "depth", "qubits", and "gateset".
        """
        # Support QubitPlaceholders: we temporarily index to arbitrary integers.
        # `generate_rb_sequence` handles mapping back to the original gateset gates.
        gateset_as_program = address_qubits(sum(gateset, Program()))
        n_qubits = len(gateset_as_program.get_qubits())
        gateset_for_api = gateset_as_program.out().splitlines()
        payload = {"depth": depth,
                   "qubits": n_qubits,
                   "gateset": gateset_for_api,
                   "seed": seed}
        return payload

    @_record_call
    def generate_rb_sequence(self, depth, gateset, seed=None):
        """
        Construct a randomized benchmarking experiment on the given qubits, decomposing into
        gateset.

        The JSON payload that is parsed is a list of lists of indices, or Nones. In the
        former case, they are the index of the gate in the gateset.

        :param int depth: The number of Clifford gates to include in the randomized benchmarking
         experiment. This is different than the number of gates in the resulting experiment.
        :param list gateset: A list of pyquil gates to decompose the Clifford elements into. These
         must generate the clifford group on the qubits of interest. e.g. for one qubit
         [RZ(np.pi/2), RX(np.pi/2)].
        :param seed: A positive integer used to seed the PRNG.
        :return: A list of pyquil programs. Each pyquil program is a circuit that represents an
         element of the Clifford group. When these programs are composed, the resulting Program
         will be the randomized benchmarking experiment of the desired depth. e.g. if the return
         programs are called cliffords then `sum(cliffords, Program())` will give the randomized
         benchmarking experiment, which will compose to the identity program.
        """
        depth = int(depth)  # needs to be jsonable, no np.int64 please!
        payload = self._rb_sequence_payload(depth, gateset, seed=seed)
        response = post_json(self.session, self.endpoint + "/rb", payload).json()
        programs = []
        for clifford in response:
            clifford_program = Program()
            # Like below, we reversed the order because the API currently hands back the Clifford
            # decomposition right-to-left.
            for index in reversed(clifford):
                clifford_program.inst(gateset[index])
            programs.append(clifford_program)
        # The programs are returned in "textbook style" right-to-left order. To compose them into
        #  the correct pyquil program, we reverse the order.
        return list(reversed(programs))


def get_benchmarker(endpoint: str = None):
    """
    Retrieve an instance of the appropriate AbstractBenchmarker subclass for a given endpoint.

    :param endpoint: Benchmarking sequence server address. Defaults to the setting in the user's
                     pyQuil config.
    :return: Instance of an AbstractBenchmarker subclass, connected to the given endpoint.
    """
    if endpoint is None:
        config = PyquilConfig()
        endpoint = config.compiler_url

    if endpoint.startswith("http"):
        return LocalBenchmarkConnection(endpoint=endpoint)
    elif endpoint.startswith("tcp"):
        return BenchmarkConnection(endpoint=endpoint)
    else:
        raise ValueError("Protocol for RB endpoint must be HTTP or TCP.")



##############################################################################
# Copyright 2016-2018 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for reading configuration information about api keys and user ids.
"""
from __future__ import print_function
from configparser import ConfigParser, NoSectionError, NoOptionError
from os.path import expanduser, abspath
from os import environ

# `.qcs_config` is for content (mostly) related to QCS: the QCS front end stacks endpoint (`url`)
# for querying all QCS data: devices, reservations, etc. etc., and the `exec_on_engage` that the
# daemon will call during an engage event
QCS_CONFIG = "QCS_CONFIG"
# `.forest_config`, for content related to the Forest SDK, such as ip addresses for the various
# servers to which users submit quil & jobs (qvm, compiler, qpu, etc.)
FOREST_CONFIG = "FOREST_CONFIG"
CONFIG_PATHS = {"QCS_CONFIG": "~/.qcs_config",
                "FOREST_CONFIG": "~/.forest_config"}


class PyquilConfig(object):
    FOREST_URL = {
        "env": "FOREST_SERVER_URL",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "url",
        "default": None
    }

    API_KEY = {
        "env": "FOREST_API_KEY",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "key",
        "default": None
    }

    USER_ID = {
        "env": "FOREST_USER_ID",
        "file": QCS_CONFIG,
        "section": "Rigetti Forest",
        "name": "user_id",
        "default": None
    }

    ENGAGE_CMD = {
        "env": "QMI_ENGAGE_CMD",
        "file": QCS_CONFIG,
        "section": "QPU",
        "name": "exec_on_engage",
        "default": ""
    }

    QPU_URL = {
        "env": "QPU_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qpu_endpoint_address",
        "default": None
    }

    QVM_URL = {
        "env": "QVM_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "qvm_address",
        "default": "http://127.0.0.1:5000"
    }

    COMPILER_URL = {
        "env": "COMPILER_URL",
        "file": FOREST_CONFIG,
        "section": "Rigetti Forest",
        "name": "compiler_server_address",
        "default": "http://127.0.0.1:6000"
    }

    def __init__(self):
        self.configparsers = {}
        for env_name, default_path in CONFIG_PATHS.items():
            default_path = expanduser(default_path)
            path = environ.get(env_name, default_path)

            cp = ConfigParser()
            cp.read(abspath(path))
            self.configparsers[env_name] = cp

    def _env_or_config_or_default(self, env=None, file=None, section=None, name=None, default=None):
        """
        Get the value of the environment variable or config file value.
        The environment variable takes precedence.

        :param env: The environment variable name.
        :param name: The config file key.
        :return: The value or None if not found
        """
        env_val = environ.get(env)
        if env_val is not None:
            return env_val

        try:
            return self.configparsers[file].get(section, name)
        except (NoSectionError, NoOptionError, KeyError):
            return default

    @property
    def api_key(self):
        return self._env_or_config_or_default(**self.API_KEY)

    @property
    def user_id(self):
        return self._env_or_config_or_default(**self.USER_ID)

    @property
    def forest_url(self):
        return self._env_or_config_or_default(**self.FOREST_URL)

    @property
    def engage_cmd(self):
        return self._env_or_config_or_default(**self.ENGAGE_CMD)

    @property
    def qpu_url(self):
        return self._env_or_config_or_default(**self.QPU_URL)

    @property
    def qvm_url(self):
        return self._env_or_config_or_default(**self.QVM_URL)

    @property
    def compiler_url(self):
        return self._env_or_config_or_default(**self.COMPILER_URL)



def wait_for_job(res, ping_time=0.5):
    raise DeprecationWarning("""
The wait_for_job function has been moved inside the QVMConnection or
QPUConnection object. For instance:
    job = qvm.wait_for_job(job_id)
    print(job.result())

See https://go.rigetti.com/connections for more info.""")


class JobResult(object):
    def __init__(self, *args, **kwargs):
        raise DeprecationWarning("JobResult has been deprecated in favor of pyquil.api.Job")



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
# THIS FILE IS DERIVED AND MODIFIED FROM PROJECTQ. COPYRIGHT PROVIDED HERE:
#
#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
##############################################################################


from copy import copy

from pyquil.latex.latex_config import get_default_settings, header, footer
from pyquil.quil import Measurement
from collections import namedtuple


command = namedtuple("command", ("gate", "lines", "ctrl_lines", "target_lines", "id"))
"""command is used as an intermediate representation to hold meta-information about the circuit, and dependencies."""
ALLOCATE = "ALLOCATE"
CZ = "CZ"
CNOT = "CNOT"
Z = "Z"
SWAP = "SWAP"
MEASURE = "MEASURE"
X = "X"


def to_latex(circuit, settings=None):
    """
    Translates a given pyquil Program to a TikZ picture in a Latex document.

    :param Program circuit: The circuit to be drawn, represented as a pyquil program.
    :param dict settings: An optional dictionary with settings for drawing the circuit. See `get_default_settings`
     in `latex_config` for more information about what settings should contain.
    :return: LaTeX document string which can be compiled.
    :rtype: string
    """
    if settings is None:
        settings = get_default_settings()
    text = header(settings)
    text += body(circuit, settings)
    text += footer()
    return text


def body(circuit, settings):
    """
    Return the body of the Latex document, including the entire circuit in
    TikZ format.

    :param Program circuit: The circuit to be drawn, represented as a pyquil program.
    :param dict settings:

    :return: Latex string to draw the entire circuit.
    :rtype: string
    """
    qubit_instruction_mapping = {}

    # Allocate each qubit.
    for inst in circuit:
        if isinstance(inst, Measurement):
            inst.qubits = [inst.qubit]
            inst.name = "MEASURE"
        else:
            qubits = inst.qubits
        for qubit in qubits:
            qubit_instruction_mapping[qubit.index] = []
    for k, v in list(qubit_instruction_mapping.items()):
        v.append(command(ALLOCATE, [k], [], [k], k))

    for inst in circuit:
        qubits = [qubit.index for qubit in inst.qubits]
        gate = inst.name
        # If this is a single qubit instruction.
        if len(qubits) == 1:
            for qubit in qubits:
                qubit_instruction_mapping[qubit].append(command(gate, [qubit], [], [qubit], qubit))

        # If this is a many-qubit operation.
        else:
            # All qubits explicitly involved in the gate.
            explicit_lines = [qubit for qubit in copy(qubits)]
            # All lines to consider that will run through this circuit element.
            all_lines = list(range(min(explicit_lines), max(explicit_lines) + 1))
            # Include all lines that are in-use and in the range of lines used in this instruction.
            for line in all_lines:
                if line not in qubit_instruction_mapping.keys() and line in all_lines:
                    all_lines.remove(line)
            for i, qubit in enumerate(all_lines):
                if gate == CZ:
                    ctrl_lines = list(explicit_lines)
                    ctrl_lines.remove(qubits[-1])
                    qubit_instruction_mapping[qubit].append(command(Z, list(all_lines), list(ctrl_lines),
                                                                    qubits[-1:], None))
                elif gate == CNOT:
                    ctrl_lines = list(explicit_lines)
                    ctrl_lines.remove(qubits[-1])
                    qubit_instruction_mapping[qubit].append(command(X, list(all_lines), list(ctrl_lines),
                                                                    qubits[-1:], None))
                else:
                    qubit_instruction_mapping[qubit].append(command(gate, list(all_lines), [],
                                                                    list(explicit_lines), None))

    # Zero index, and remove gaps in spacing.
    relabeled_circuit = {}
    # Store a mapping so we can relabel command labels.
    index_map = {}
    for i, key in enumerate(sorted(qubit_instruction_mapping.keys())):
        relabeled_circuit[i] = qubit_instruction_mapping[key]
        index_map[key] = i

    for line in list(relabeled_circuit.values()):
        for cmd in line:
            for i, qubit in enumerate(cmd.lines):
                cmd.lines[i] = index_map[qubit]
            for i, qubit in enumerate(cmd.ctrl_lines):
                cmd.ctrl_lines[i] = index_map[qubit]
            for i, qubit in enumerate(cmd.target_lines):
                cmd.target_lines[i] = index_map[qubit]
    code_generator = CircuitTikzGenerator(settings)
    return code_generator.generate_circuit(relabeled_circuit)


class CircuitTikzGenerator(object):
    """
    Takes a circuit and turns it into Latex/TikZ code.
    """
    def __init__(self, settings):
        """
        Initialize a circuit to LaTeX converter object.

        :param dict settings: Dictionary of settings to use for the TikZ image.
        :param int num_lines: Number of qubit lines to use for the entire circuit.
        """
        self.pos = None
        self.op_count = None
        self.is_quantum = None
        self.settings = settings

    def generate_circuit(self, circuit_description):

        num_lines = len(circuit_description)
        self.pos = [0.] * num_lines
        self.op_count = [0] * num_lines
        self.is_quantum = [self.settings['lines']['init_quantum']] * num_lines
        code = []
        for line in circuit_description.keys():
            code.append(self.to_tikz(line, circuit_description))
        return "".join(code)

    def to_tikz(self, line, circuit, end=None):
        """
        Generate the TikZ code for one line of the circuit up to a certain
        gate.

        It modifies the circuit to include only the gates which have not been
        drawn. It automatically switches to other lines if the gates on those
        lines have to be drawn earlier.

        :param int line: Line to generate the TikZ code for.
        :param Program circuit: The circuit to draw.
        :param int end: Gate index to stop at (for recursion).
        :return:TikZ code representing the current qubit line and, if it was
         necessary to draw other lines, those lines as well.
        :rtype: string
        """
        if end is None:
            end = len(circuit[line])
        tikz_code = []
        cmds = circuit[line]
        for i in range(0, end):
            gate = cmds[i].gate
            lines = cmds[i].lines
            ctrl_lines = cmds[i].ctrl_lines
            all_lines = list(lines)
            all_lines.remove(line)
            for l in all_lines:
                gate_idx = 0
                while not (circuit[l][gate_idx] == cmds[i]):
                    gate_idx += 1
                tikz_code.append(self.to_tikz(l, circuit, gate_idx))

                # We remove the first element of the current line, since we have processed it.
                circuit[l] = circuit[l][1:]
            all_lines = cmds[i].target_lines + cmds[i].ctrl_lines
            pos = max([self.pos[l] for l in range(min(all_lines),
                                                  max(all_lines) + 1)])
            for l in range(min(all_lines), max(all_lines) + 1):
                self.pos[l] = pos + self._gate_pre_offset(gate)

            connections = ""
            for l in all_lines:
                connections += self._line(self.op_count[l] - 1, self.op_count[l], line=l)
            add_str = ""
            if gate == X:
                # draw NOT-gate with controls
                add_str = self._x_gate(cmds[i].target_lines, ctrl_lines)
                # and make the target qubit quantum if one of the controls is
                if not self.is_quantum[lines[0]]:
                    if sum([self.is_quantum[i] for i in ctrl_lines]) > 0:
                        self.is_quantum[lines[0]] = True
            elif gate == Z and len(ctrl_lines) > 0:
                add_str = self._cz_gate(lines + ctrl_lines)
            elif gate == SWAP:
                add_str = self._swap_gate(lines, ctrl_lines)
            elif gate == MEASURE:
                # draw measurement gate
                for l in lines:
                    op = self._op(l)
                    width = self._gate_width(MEASURE)
                    height = self._gate_height(MEASURE)
                    shift0 = .07 * height
                    shift1 = .36 * height
                    shift2 = .1 * width
                    add_str += ("\n\\node[measure,edgestyle] ({op}) at ({pos},-{line}) {{}};\n"
                                "\\draw[edgestyle] ([yshift=-{shift1}cm,xshift={shift2}cm]{op}.west)"
                                " to [out=60,in=180] ([yshift={shift0}cm]{op}."
                                "center) to [out=0, in=120] ([yshift=-{shift1}"
                                "cm,xshift=-{shift2}cm]{op}.east);\n"
                                "\\draw[edgestyle] ([yshift=-{shift1}cm]{op}."
                                "center) to ([yshift=-{shift2}cm,xshift=-{shift1}cm]{op}.north east);").format(
                        op=op, pos=self.pos[l], line=l, shift0=shift0, shift1=shift1, shift2=shift2)
                    self.op_count[l] += 1
                    self.pos[l] += (self._gate_width(gate) + self._gate_offset(gate))
                    self.is_quantum[l] = False
            elif gate == ALLOCATE:
                # Draw 'begin line'.
                add_str = "\n\\node[none] ({}) at ({},-{}) {{$\\Ket{{0}}{}$}};"
                id_str = ""
                if self.settings['gates']['AllocateQubitGate']['draw_id']:
                    id_str = "^{{\\textcolor{{red}}{{{}}}}}".format(cmds[i].id)
                add_str = add_str.format(self._op(line), self.pos[line], line,
                                         id_str)
                self.op_count[line] += 1
                self.pos[line] += (self._gate_offset(gate) + self._gate_width(gate))
                self.is_quantum[line] = self.settings['lines']['init_quantum']
            else:
                # Regular gate must draw the lines it does not act upon.
                # If it spans multiple qubits.
                add_str = self._regular_gate(gate, lines, ctrl_lines, cmds[i].target_lines)
                for l in lines:
                    self.is_quantum[l] = True
            tikz_code.append(add_str)
            if not gate == ALLOCATE:
                tikz_code.append(connections)
        circuit[line] = circuit[line][end:]
        return "".join(tikz_code)

    def _gate_name(self, gate):
        """
        Return the string representation of the gate.

        Tries to use gate.tex_str and, if that is not available, uses str(gate)
        instead.

        :param string gate: Gate object of which to get the name / LaTeX representation.
        :return: LaTeX gate name.
        :rtype: string
        """
        try:
            name = gate.tex_str()
        except AttributeError:
            name = str(gate)
        return name

    def _swap_gate(self, lines, ctrl_lines):
        """
        Return the TikZ code for a Swap-gate.

        :param lines: List of length 2 denoting the target qubit of the Swap gate.
        :type: list[int]
        :param ctrl_lines: List of qubit lines which act as controls.
        :type: list[int]
        """
        delta_pos = self._gate_offset(SWAP)
        gate_width = self._gate_width(SWAP)
        lines.sort()

        gate_str = ""
        for line in lines:
            op = self._op(line)
            w = "{}cm".format(.5 * gate_width)
            s1 = "[xshift=-{w},yshift=-{w}]{op}.center".format(w=w, op=op)
            s2 = "[xshift={w},yshift={w}]{op}.center".format(w=w, op=op)
            s3 = "[xshift=-{w},yshift={w}]{op}.center".format(w=w, op=op)
            s4 = "[xshift={w},yshift=-{w}]{op}.center".format(w=w, op=op)
            swap_style = "swapstyle,edgestyle"
            if self.settings['gate_shadow']:
                swap_style += ",shadowed"
            gate_str += ("\n\\node[swapstyle] ({op}) at ({pos},-{line}) {{}};"
                         "\n\\draw[{swap_style}] ({s1})--({s2});\n"
                         "\\draw[{swap_style}] ({s3})--({s4});"
                         ).format(op=op, s1=s1, s2=s2, s3=s3, s4=s4,
                                  line=line, pos=self.pos[line],
                                  swap_style=swap_style)
        gate_str += self._line(lines[0], lines[1])
        all_lines = ctrl_lines + lines
        new_pos = self.pos[lines[0]] + delta_pos + gate_width
        for i in all_lines:
            self.op_count[i] += 1
        for i in range(min(all_lines), max(all_lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _x_gate(self, lines, ctrl_lines):
        """
        Return the TikZ code for a NOT-gate.

        :param lines: List of length 1 denoting the target qubit of the NOT / X gate.
        :type: list[int]
        :param ctrl_lines: List of qubit lines which act as controls.
        :type: list[int]
        """
        line = lines[0]
        delta_pos = self._gate_offset('X')
        gate_width = self._gate_width('X')
        op = self._op(line)
        gate_str = ("\n\\node[xstyle] ({op}) at ({pos},-{line}) {{}};\n\\draw"
                    "[edgestyle] ({op}.north)--({op}.south);\n\\draw"
                    "[edgestyle] ({op}.west)--({op}.east);").format(op=op, line=line, pos=self.pos[line])
        if len(ctrl_lines) > 0:
            for ctrl in ctrl_lines:
                gate_str += self._phase(ctrl, self.pos[line])
                gate_str += self._line(ctrl, line)

        all_lines = ctrl_lines + [line]
        new_pos = self.pos[line] + delta_pos + gate_width
        for i in all_lines:
            self.op_count[i] += 1
        for i in range(min(all_lines), max(all_lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _cz_gate(self, lines):
        """
        Return the TikZ code for an n-controlled Z-gate.

        :param lines: List of all qubits involved.
        :type: list[int]
        """
        line = lines[0]
        delta_pos = self._gate_offset(Z)
        gate_width = self._gate_width(Z)
        gate_str = self._phase(line, self.pos[line])

        for ctrl in lines[1:]:
            gate_str += self._phase(ctrl, self.pos[line])
            gate_str += self._line(ctrl, line)

        new_pos = self.pos[line] + delta_pos + gate_width
        for i in lines:
            self.op_count[i] += 1
        for i in range(min(lines), max(lines) + 1):
            self.pos[i] = new_pos
        return gate_str

    def _gate_width(self, gate):
        """
        Return the gate width, using the settings (if available).

        :param string gate: The name of the gate whose height is desired.
        :return: Width of the gate.
        :rtype: float
        """
        try:
            gates = self.settings['gates']
            gate_width = gates[gate.__class__.__name__]['width']
        except KeyError:
            gate_width = .5
        return gate_width

    def _gate_pre_offset(self, gate):
        """
        Return the offset to use before placing this gate.

        :param string gate: The name of the gate whose pre-offset is desired.
        :return: Offset to use before the gate.
        :rtype: float
        """
        try:
            gates = self.settings['gates']
            delta_pos = gates[gate.__class__.__name__]['pre_offset']
        except KeyError:
            delta_pos = self._gate_offset(gate)
        return delta_pos

    def _gate_offset(self, gate):
        """
        Return the offset to use after placing this gate and, if no pre_offset
        is defined, the same offset is used in front of the gate.

        :param string gate: The name of the gate whose offset is desired.
        :return: Offset.
        :rtype: float
        """
        try:
            gates = self.settings['gates']
            delta_pos = gates[gate.__class__.__name__]['offset']
        except KeyError:
            delta_pos = .2
        return delta_pos

    def _gate_height(self, gate):
        """
        Return the height to use for this gate.

        :param string gate: The name of the gate whose height is desired.
        :return: Height of the gate.
        :rtype: float
        """
        try:
            height = self.settings['gates'][gate.__class__.__name__]['height']
        except KeyError:
            height = .5
        return height

    def _phase(self, line, pos):
        """
        Places a phase / control circle on a qubit line at a given position.

        :param int line: Qubit line at which to place the circle.
        :param float pos: Position at which to place the circle.
        :return: Latex string representing a control circle at the given position.
        :rtype: string
        """
        phase_str = "\n\\node[phase] ({}) at ({},-{}) {{}};"
        return phase_str.format(self._op(line), pos, line)

    def _op(self, line, op=None, offset=0):
        """
        Returns the gate name for placing a gate on a line.

        :param int line: Line number.
        :param int op: Operation number or, by default, uses the current op count.
        :return: Gate name.
        :rtype: string
        """
        if op is None:
            op = self.op_count[line]
        return "line{}_gate{}".format(line, op + offset)

    def _line(self, p1, p2, line=None):
        """
        Connects p1 and p2, where p1 and p2 are either to qubit line indices,
        in which case the two most recent gates are connected, or two gate
        indices, in which case line denotes the line number and the two gates
        are connected on the given line.

        :param int p1: Index of the first object to connect.
        :param int p2: Index of the second object to connect.
        :param int line: Line index - if provided, p1 and p2 are gate indices.
        :return: Latex code to draw this / these line(s).
        :rtype: string
        """
        dbl_classical = self.settings['lines']['double_classical']

        if line is None:
            quantum = not dbl_classical or self.is_quantum[p1]
            op1, op2 = self._op(p1), self._op(p2)
            loc1, loc2 = 'north', 'south'
            shift = "xshift={}cm"
        else:
            quantum = not dbl_classical or self.is_quantum[line]
            op1, op2 = self._op(line, p1), self._op(line, p2)
            loc1, loc2 = 'west', 'east'
            shift = "yshift={}cm"

        if quantum:
            return "\n\\draw ({}) edge[edgestyle] ({});".format(op1, op2)
        else:
            if p2 > p1:
                loc1, loc2 = loc2, loc1
            edge_str = "\n\\draw ([{shift}]{op1}.{loc1}) edge[edgestyle] ([{shift}]{op2}.{loc2});"
            line_sep = self.settings['lines']['double_lines_sep']
            shift1 = shift.format(line_sep / 2.)
            shift2 = shift.format(-line_sep / 2.)
            edges_str = edge_str.format(shift=shift1, op1=op1, op2=op2, loc1=loc1, loc2=loc2)
            edges_str += edge_str.format(shift=shift2, op1=op1, op2=op2, loc1=loc1, loc2=loc2)
            return edges_str

    def _regular_gate(self, gate, lines, ctrl_lines, used_lines):
        """
        Draw a regular gate.

        :param string gate: Gate to draw.
        :param lines: Lines the gate acts on.
        :type: list[int]
        :param int ctrl_lines: Control lines.
        :param int used_lines: The lines that are actually involved in the gate.
        :return: LaTeX string drawing a regular gate at the given location.
        :rtype: string
        """
        imax = max(lines)
        imin = min(lines)

        delta_pos = self._gate_offset(gate)
        gate_width = self._gate_width(gate)
        gate_height = self._gate_height(gate)

        name = self._gate_name(gate)

        lines = list(range(imin, imax + 1))

        tex_str = ""
        pos = self.pos[lines[0]]

        node_str = "\n\\node[none] ({}) at ({},-{}) {{}};"
        for l in lines:
            node1 = node_str.format(self._op(l), pos, l)
            if l in used_lines:
                tex_str += self._phase(l, pos)
            node2 = ("\n\\node[none,minimum height={}cm,outer sep=0] ({}) at"
                     " ({},-{}) {{}};").format(gate_height, self._op(l, offset=1), pos + gate_width / 2., l)
            node3 = node_str.format(self._op(l, offset=2), pos + gate_width, l)
            tex_str += node1 + node2 + node3
        tex_str += ("\n\\draw[operator,edgestyle,outer sep={width}cm]"
                    " ([yshift={half_height}cm]{op1})"
                    " rectangle ([yshift=-{half_height}cm]{op2}) node[pos=.5]{{\\verb|{name}|}};"
                    ).format(width=gate_width, op1=self._op(imin), op2=self._op(imax, offset=2),
                             half_height=.5 * gate_height, name=name)
        for l in lines:
            self.pos[l] = pos + gate_width / 2.
            self.op_count[l] += 3
        for l in range(min(ctrl_lines + lines), max(ctrl_lines + lines) + 1):
            self.pos[l] = pos + delta_pos + gate_width
        return tex_str
